
import pytest
import decimal

import pandas as pd
import numpy as np

from rcv_cruncher.cvr.base import CastVoteRecord
from rcv_cruncher.marks import BallotMarks


add_rule_set_ballots = [
    [
        'A',
        BallotMarks.WRITEIN,
        'write-in',
        BallotMarks.SKIPPED,
        BallotMarks.SKIPPED,
        BallotMarks.OVERVOTE,
        'A',
        'B'
    ],
    [
        BallotMarks.OVERVOTE,
        'Tuwi',
        BallotMarks.WRITEIN,
        'A',
        'B',
        'B',
        'C',
        BallotMarks.SKIPPED
    ]
]

params = [
    ({
        'input': {
            'cvr': {
                'weight': [1, 1],
                'ranks': add_rule_set_ballots
            },
            'rule_set': BallotMarks.new_rule_set(
                combine_writein_marks=True
            )
        },
        'expected': {
            'cvr': {
                'weight': [decimal.Decimal('1'), decimal.Decimal('1')],
                'ranks': [
                    [
                        'A',
                        BallotMarks.WRITEIN,
                        BallotMarks.WRITEIN,  # 'write-in',
                        BallotMarks.SKIPPED,
                        BallotMarks.SKIPPED,
                        BallotMarks.OVERVOTE,
                        'A',
                        'B'
                    ],
                    [
                        BallotMarks.OVERVOTE,
                        'Tuwi',
                        BallotMarks.WRITEIN,
                        'A',
                        'B',
                        'B',
                        'C',
                        BallotMarks.SKIPPED
                    ]
                ]
            },
            'candidate_set': {
                'A',
                'B',
                'C',
                'Tuwi',
                # 'write-in',
                BallotMarks.WRITEIN
            }
        }
    }),
    ({
        'input': {
            'cvr': {
                'weight': [1, 1],
                'ranks': add_rule_set_ballots
            },
            'rule_set': BallotMarks.new_rule_set(
                combine_writein_marks=True,
                exclude_writein_marks=True
            )
        },
        'expected': {
            'cvr': {
                'weight': [decimal.Decimal('1'), decimal.Decimal('1')],
                'ranks': [
                    [
                        'A',
                        # BallotMarks.WRITEIN,
                        # 'write-in',
                        BallotMarks.SKIPPED,
                        BallotMarks.SKIPPED,
                        BallotMarks.OVERVOTE,
                        'A',
                        'B'
                    ],
                    [
                        BallotMarks.OVERVOTE,
                        'Tuwi',
                        # BallotMarks.WRITEIN,
                        'A',
                        'B',
                        'B',
                        'C',
                        BallotMarks.SKIPPED
                    ]
                ]
            },
            'candidate_set': {
                'A',
                'B',
                'C',
                'Tuwi'
                # 'write-in',
                # BallotMarks.WRITEIN
            }
        }
    }),
    ({
        'input': {
            'cvr': {
                'weight': [1, 1],
                'ranks': add_rule_set_ballots
            },
            'rule_set': BallotMarks.new_rule_set(
                combine_writein_marks=False,
                exclude_writein_marks=True
            )
        },
        'expected': {
            'cvr': {
                'weight': [decimal.Decimal('1'), decimal.Decimal('1')],
                'ranks': [
                    [
                        'A',
                        # BallotMarks.WRITEIN,
                        'write-in',
                        BallotMarks.SKIPPED,
                        BallotMarks.SKIPPED,
                        BallotMarks.OVERVOTE,
                        'A',
                        'B'
                    ],
                    [
                        BallotMarks.OVERVOTE,
                        'Tuwi',
                        # BallotMarks.WRITEIN,
                        'A',
                        'B',
                        'B',
                        'C',
                        BallotMarks.SKIPPED
                    ]
                ]
            },
            'candidate_set': {
                'A',
                'B',
                'C',
                'Tuwi',
                'write-in',
                # BallotMarks.WRITEIN
            }
        }
    }),
    ({
        'input': {
            'cvr': {
                'weight': [1, 1],
                'ranks': add_rule_set_ballots
            },
            'rule_set': BallotMarks.new_rule_set(
                exclude_duplicate_candidate_marks=True
            )
        },
        'expected': {
            'cvr': {
                'weight': [decimal.Decimal('1'), decimal.Decimal('1')],
                'ranks': [
                    [
                        'A',
                        BallotMarks.WRITEIN,
                        'write-in',
                        BallotMarks.SKIPPED,
                        BallotMarks.SKIPPED,
                        BallotMarks.OVERVOTE,
                        # 'A',
                        'B'
                    ],
                    [
                        BallotMarks.OVERVOTE,
                        'Tuwi',
                        BallotMarks.WRITEIN,
                        'A',
                        'B',
                        # 'B',
                        'C',
                        BallotMarks.SKIPPED
                    ]
                ]
            },
            'candidate_set': {
                'A',
                'B',
                'C',
                'Tuwi',
                'write-in',
                BallotMarks.WRITEIN
            }
        }
    }),
    ({
        'input': {
            'cvr': {
                'weight': [1, 1],
                'ranks': add_rule_set_ballots
            },
            'rule_set': BallotMarks.new_rule_set(
                exclude_overvote_marks=True
            )
        },
        'expected': {
            'cvr': {
                'weight': [decimal.Decimal('1'), decimal.Decimal('1')],
                'ranks': [
                    [
                        'A',
                        BallotMarks.WRITEIN,
                        'write-in',
                        BallotMarks.SKIPPED,
                        BallotMarks.SKIPPED,
                        # BallotMarks.OVERVOTE,
                        'A',
                        'B'
                    ],
                    [
                        # BallotMarks.OVERVOTE,
                        'Tuwi',
                        BallotMarks.WRITEIN,
                        'A',
                        'B',
                        'B',
                        'C',
                        BallotMarks.SKIPPED
                    ]
                ]
            },
            'candidate_set': {
                'A',
                'B',
                'C',
                'Tuwi',
                'write-in',
                BallotMarks.WRITEIN
            }
        }
    }),
    ({
        'input': {
            'cvr': {
                'weight': [1, 1],
                'ranks': add_rule_set_ballots
            },
            'rule_set': BallotMarks.new_rule_set(
                exclude_skipped_marks=True
            )
        },
        'expected': {
            'cvr': {
                'weight': [decimal.Decimal('1'), decimal.Decimal('1')],
                'ranks': [
                    [
                        'A',
                        BallotMarks.WRITEIN,
                        'write-in',
                        # BallotMarks.SKIPPED,
                        # BallotMarks.SKIPPED,
                        BallotMarks.OVERVOTE,
                        'A',
                        'B'
                    ],
                    [
                        BallotMarks.OVERVOTE,
                        'Tuwi',
                        BallotMarks.WRITEIN,
                        'A',
                        'B',
                        'B',
                        'C',
                        # BallotMarks.SKIPPED
                    ]
                ]
            },
            'candidate_set': {
                'A',
                'B',
                'C',
                'Tuwi',
                'write-in',
                BallotMarks.WRITEIN
            }
        }
    }),
    ({
        'input': {
            'cvr': {
                'weight': [1, 1],
                'ranks': add_rule_set_ballots
            },
            'rule_set': BallotMarks.new_rule_set(
                exhaust_on_overvote_marks=True
            )
        },
        'expected': {
            'cvr': {
                'weight': [decimal.Decimal('1'), decimal.Decimal('1')],
                'ranks': [
                    [
                        'A',
                        BallotMarks.WRITEIN,
                        'write-in',
                        BallotMarks.SKIPPED,
                        BallotMarks.SKIPPED,
                        # BallotMarks.OVERVOTE,
                        # 'A',
                        # 'B'
                    ],
                    [
                        # BallotMarks.OVERVOTE,
                        # 'Tuwi',
                        # BallotMarks.WRITEIN,
                        # 'A',
                        # 'B',
                        # 'B',
                        # 'C',
                        # BallotMarks.SKIPPED
                    ]
                ]
            },
            'candidate_set': {
                'A',
                'B',
                'C',
                'Tuwi',
                'write-in',
                BallotMarks.WRITEIN
            }
        }
    }),
    ({
        'input': {
            'cvr': {
                'weight': [1, 1],
                'ranks': add_rule_set_ballots
            },
            'rule_set': BallotMarks.new_rule_set(
                exhaust_on_repeated_skipped_marks=True
            )
        },
        'expected': {
            'cvr': {
                'weight': [decimal.Decimal('1'), decimal.Decimal('1')],
                'ranks': [
                    [
                        'A',
                        BallotMarks.WRITEIN,
                        'write-in',
                        # BallotMarks.SKIPPED,
                        # BallotMarks.SKIPPED,
                        # BallotMarks.OVERVOTE,
                        # 'A',
                        # 'B'
                    ],
                    [
                        BallotMarks.OVERVOTE,
                        'Tuwi',
                        BallotMarks.WRITEIN,
                        'A',
                        'B',
                        'B',
                        'C',
                        BallotMarks.SKIPPED
                    ]
                ]
            },
            'candidate_set': {
                'A',
                'B',
                'C',
                'Tuwi',
                'write-in',
                BallotMarks.WRITEIN
            }
        }
    }),
    ({
        'input': {
            'cvr': {
                'weight': [1, 1],
                'ranks': add_rule_set_ballots
            },
            'rule_set': BallotMarks.new_rule_set(
                exhaust_on_duplicate_candidate_marks=True
            )
        },
        'expected': {
            'cvr': {
                'weight': [decimal.Decimal('1'), decimal.Decimal('1')],
                'ranks': [
                    [
                        'A',
                        BallotMarks.WRITEIN,
                        'write-in',
                        BallotMarks.SKIPPED,
                        BallotMarks.SKIPPED,
                        BallotMarks.OVERVOTE,
                        # 'A',
                        # 'B'
                    ],
                    [
                        BallotMarks.OVERVOTE,
                        'Tuwi',
                        BallotMarks.WRITEIN,
                        'A',
                        'B',
                        # 'B',
                        # 'C',
                        # BallotMarks.SKIPPED
                    ]
                ]
            },
            'candidate_set': {
                'A',
                'B',
                'C',
                'Tuwi',
                'write-in',
                BallotMarks.WRITEIN
            }
        }
    }),
    ({
        'input': {
            'cvr': {
                'weight': [1, 1],
                'ranks': add_rule_set_ballots
            },
            'rule_set': BallotMarks.new_rule_set(
                exhaust_on_duplicate_candidate_marks=True,
                treat_combined_writeins_as_exhaustable_duplicates=True,
                combine_writein_marks=True
            )
        },
        'expected': {
            'cvr': {
                'weight': [decimal.Decimal('1'), decimal.Decimal('1')],
                'ranks': [
                    [
                        'A',
                        BallotMarks.WRITEIN,
                        # 'write-in',
                        # BallotMarks.SKIPPED,
                        # BallotMarks.SKIPPED,
                        # BallotMarks.OVERVOTE,
                        # 'A',
                        # 'B'
                    ],
                    [
                        BallotMarks.OVERVOTE,
                        'Tuwi',
                        BallotMarks.WRITEIN,
                        'A',
                        'B',
                        # 'B',
                        # 'C',
                        # BallotMarks.SKIPPED
                    ]
                ]
            },
            'candidate_set': {
                'A',
                'B',
                'C',
                'Tuwi',
                # 'write-in',
                BallotMarks.WRITEIN
            }
        }
    })
]


@pytest.mark.parametrize("param", params)
def test_add_rule_set(param):

    cast_vote_record = CastVoteRecord(parsed_cvr=param['input']['cvr'])
    cast_vote_record.add_rule_set('test', param['input']['rule_set'])

    cvr_dict = cast_vote_record.get_cvr_dict('test')
    cvr_dict['ranks'] = [b.marks for b in cvr_dict['ballot_marks']]
    del cvr_dict['ballot_marks']

    candidate_set = set(cast_vote_record.get_candidates('test').marks)

    assert cvr_dict == param['expected']['cvr']
    assert candidate_set == param['expected']['candidate_set']


params = [
    ({
        'input': {
            'cvr': {
                'ranks': add_rule_set_ballots
            },
            'format': 'rank'
        },
        'expected': {
            'table': pd.DataFrame({
                'rank1': ['A', BallotMarks.OVERVOTE],
                'rank2': [BallotMarks.WRITEIN, 'Tuwi'],
                'rank3': ['write-in', BallotMarks.WRITEIN],
                'rank4': [BallotMarks.SKIPPED, 'A'],
                'rank5': [BallotMarks.SKIPPED, 'B'],
                'rank6': [BallotMarks.OVERVOTE, 'B'],
                'rank7': ['A', 'C'],
                'rank8': ['B', BallotMarks.SKIPPED]
            })
        }
    }),
    ({
        'input': {
            'cvr': {
                'ranks': add_rule_set_ballots
            },
            'format': 'candidate'
        },
        'expected': {
            'table': pd.DataFrame({
                'candidate_A': ['1,7', '4'],
                'candidate_B': ['8', '5,6'],
                'candidate_C': [np.nan, '7'],
                'candidate_Tuwi': [np.nan, '2'],
                'candidate_write-in': ['3', np.nan],
                f'candidate_{BallotMarks.WRITEIN}': ['2', '3'],
                f'candidate_{BallotMarks.OVERVOTE}': ['6', '1'],
                'rank_limit': [8, 8]
            })
        }
    })
]


@pytest.mark.parametrize("param", params)
def test_get_cvr_table(param):

    cast_vote_record = CastVoteRecord(parsed_cvr=param['input']['cvr'])
    table = cast_vote_record.cvr_table(table_format=param['input']['format'])

    for col in table.columns:
        assert table[col].equals(param['expected']['table'][col])


params = [
    ({
        'expected': {
            'stat': 1,
            'split_stat': [1]
        },
        'input': {
            'cvr': {
                'ranks': [['A']],
                'weight': [1]
            }
        }
    }),
    ({
        'expected': {
            'stat': 2
        },
        'input': {
            'cvr': {
                'ranks': [['A', 'B']],
                'weight': [1]
            }
        }
    }),
    ({
        'expected': {
            'stat': 3
        },
        'input': {
            'cvr': {
                'ranks': [['A', 'B', 'C']],
                'weight': [1]
            }
        }
    }),
    ({
        'expected': {
            'stat': 1
        },
        'input': {
            'cvr': {
                'ranks': [
                    ['A'],
                    [BallotMarks.SKIPPED]
                ],
                'weight': [1, 1]
            }
        }
    }),
    ({
        'expected': {
            'stat': 2
        },
        'input': {
            'cvr': {
                'ranks': [
                    ['A', 'B'],
                    [BallotMarks.OVERVOTE, 'C']
                ],
                'weight': [1, 1]
            }
        }
    }),
    ({
        'expected': {
            'stat': 3
        },
        'input': {
            'cvr': {
                'ranks': [
                    ['A', 'B', 'C'],
                    [BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.SKIPPED]
                ],
                'weight': [1, 1]
            }
        }
    })
]


@pytest.mark.parametrize("param", params)
def test_rank_limit(param):
    cast_vote_record = CastVoteRecord(parsed_cvr=param['input']['cvr'])

    computed_stat = cast_vote_record.stats()['rank_limit'].item()
    assert param['expected']['stat'] == computed_stat


params = [
    ({
        'expected': {
            'stat': False
        },
        'input': {
            'cvr': {
                'ranks': [['A', 'B', 'C']],
                'weight': [1]
            }
        }
    }),
    ({
        'expected': {
            'stat': False
        },
        'input': {
            'cvr': {
                'ranks': [
                    ['A', 'B', 'C'],
                    [BallotMarks.SKIPPED, BallotMarks.SKIPPED, 'D']
                    ],
                'weight': [1, 1]
            }
        }
    }),
    ({
        'expected': {
            'stat': True
        },
        'input': {
            'cvr': {
                'ranks': [
                    ['A', 'B', 'C'],
                    [BallotMarks.SKIPPED, 'E', 'D']
                    ],
                'weight': [1, 1]
            }
        }
    }),
]


@pytest.mark.parametrize("param", params)
def test_restrictive_rank_limit(param):
    cast_vote_record = CastVoteRecord(parsed_cvr=param['input']['cvr'])
    computed_stat = cast_vote_record.stats()['restrictive_rank_limit'].item()
    assert param['expected']['stat'] == computed_stat


params = [
    ({
        'expected': {
            'stat': 0
        },
        'input': {
            'cvr': {
                'ranks': [[BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.OVERVOTE]],
                'weight': [1]
            }
        }
    }),
    ({
        'expected': {
            'stat': 0
        },
        'input': {
            'cvr': {
                'ranks': [
                    [BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.OVERVOTE],
                    [BallotMarks.SKIPPED, BallotMarks.WRITEIN, BallotMarks.SKIPPED]
                    ],
                'weight': [1, 1]
            }
        }
    }),
    ({
        'expected': {
            'stat': 1
        },
        'input': {
            'cvr': {
                'ranks': [
                    [BallotMarks.SKIPPED, 'A', BallotMarks.OVERVOTE],
                    [BallotMarks.SKIPPED, BallotMarks.WRITEIN, BallotMarks.SKIPPED]
                    ],
                'weight': [1, 1]
            }
        }
    }),
    ({
        'expected': {
            'stat': 2
        },
        'input': {
            'cvr': {
                'ranks': [
                    [BallotMarks.SKIPPED, 'A', BallotMarks.OVERVOTE],
                    [BallotMarks.SKIPPED, BallotMarks.WRITEIN, 'B']
                    ],
                'weight': [1, 1]
            }
        }
    }),
]


@pytest.mark.parametrize("param", params)
def test_n_candidates(param):
    cast_vote_record = CastVoteRecord(parsed_cvr=param['input']['cvr'])
    computed_stat = cast_vote_record.stats()['n_candidates'].item()
    assert param['expected']['stat'] == computed_stat


params = [
    ({
        'expected': {
            'stat': 0
        },
        'input': {
            'cvr': {
                'ranks': [[BallotMarks.SKIPPED, 'A', BallotMarks.OVERVOTE]],
                'weight': [1]
            }
        }
    }),
    ({
        'expected': {
            'stat': 1
        },
        'input': {
            'cvr': {
                'ranks': [[BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.OVERVOTE]],
                'weight': [1]
            }
        }
    }),
    ({
        'expected': {
            'stat': 2
        },
        'input': {
            'cvr': {
                'ranks': [
                    [BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.OVERVOTE],
                    [BallotMarks.OVERVOTE, 'A', BallotMarks.OVERVOTE]
                    ],
                'weight': [1, 1]
            }
        }
    })
]


@pytest.mark.parametrize("param", params)
def test_first_round_overvote(param):
    cast_vote_record = CastVoteRecord(parsed_cvr=param['input']['cvr'])
    computed_stat = cast_vote_record.stats()['first_round_overvote'].item()
    assert param['expected']['stat'] == computed_stat


params = [
    ({
        'expected': {
            'stat': [0]
        },
        'input': {
            'cvr': {
                'ranks': [['A']],
                'weight': [1],
                'split': [1]
            }
        }
    }),
    ({
        'expected': {
            'stat': [0, 0]
        },
        'input': {
            'cvr': {
                'ranks': [
                    ['A', 'B'],
                    [BallotMarks.SKIPPED, 'A']
                ],
                'weight': [1, 1],
                'split': [1, 2]
            }
        }
    }),
    ({
        'expected': {
            'stat': [0, 2]
        },
        'input': {
            'cvr': {
                'ranks': [
                    ['A', 'B'],
                    [BallotMarks.SKIPPED, BallotMarks.OVERVOTE],
                    [BallotMarks.OVERVOTE, 'A']
                ],
                'weight': [1, 1, 1],
                'split': [1, 2, 2]
            }
        }
    }),
]


@pytest.mark.parametrize("param", params)
def test_split_first_round_overvote(param):
    cast_vote_record = CastVoteRecord(parsed_cvr=param['input']['cvr'], split_fields=['split'])
    computed_stat = cast_vote_record.stats(add_split_stats=True)['split_first_round_overvote'].tolist()
    assert param['expected']['stat'] == computed_stat


params = [
    ({
        'expected': {
            'stat': 1
        },
        'input': {
            'cvr': {
                'ranks': [[BallotMarks.SKIPPED, 'A', BallotMarks.OVERVOTE]],
                'weight': [1]
            }
        }
    }),
    ({
        'expected': {
            'stat': 0
        },
        'input': {
            'cvr': {
                'ranks': [[BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.OVERVOTE]],
                'weight': [1]
            }
        }
    }),
    ({
        'expected': {
            'stat': 0
        },
        'input': {
            'cvr': {
                'ranks': [
                    [BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.OVERVOTE],
                    [BallotMarks.SKIPPED, BallotMarks.WRITEIN, 'B']
                ],
                'weight': [1, 1]
            }
        }
    }),
    ({
        'expected': {
            'stat': 5
        },
        'input': {
            'cvr': {
                'ranks': [
                    [BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.OVERVOTE],
                    [BallotMarks.SKIPPED, BallotMarks.SKIPPED, 'B']
                ],
                'weight': [1, 5]
            }
        }
    }),
]


@pytest.mark.parametrize("param", params)
def test_ranked_single(param):
    cast_vote_record = CastVoteRecord(parsed_cvr=param['input']['cvr'])
    computed_stat = cast_vote_record.stats()['ranked_single'].item()
    assert param['expected']['stat'] == computed_stat


params = [
    ({
        'expected': {
            'stat': [1]
        },
        'input': {
            'cvr': {
                'ranks': [['A']],
                'weight': [1],
                'split': [1]
            }
        }
    }),
    ({
        'expected': {
            'stat': [0, 1]
        },
        'input': {
            'cvr': {
                'ranks': [
                    ['A', 'B'],
                    [BallotMarks.SKIPPED, 'A']
                ],
                'weight': [1, 1],
                'split': [1, 2]
            }
        }
    }),
    ({
        'expected': {
            'stat': [0, 1]
        },
        'input': {
            'cvr': {
                'ranks': [
                    ['A', 'B'],
                    [BallotMarks.SKIPPED, BallotMarks.OVERVOTE],
                    [BallotMarks.OVERVOTE, 'A']
                ],
                'weight': [1, 1, 1],
                'split': [1, 2, 2]
            }
        }
    }),
]


@pytest.mark.parametrize("param", params)
def test_split_ranked_single(param):
    cast_vote_record = CastVoteRecord(parsed_cvr=param['input']['cvr'], split_fields=['split'])
    computed_stat = cast_vote_record.stats(add_split_stats=True)['split_ranked_single'].tolist()
    assert param['expected']['stat'] == computed_stat


params = [
    ({
        'expected': {
            'stat': 0
        },
        'input': {
            'cvr': {
                'ranks': [[BallotMarks.SKIPPED, 'A', BallotMarks.OVERVOTE]],
                'weight': [1]
            }
        }
    }),
    ({
        'expected': {
            'stat': 0
        },
        'input': {
            'cvr': {
                'ranks': [[BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.OVERVOTE]],
                'weight': [1]
            }
        }
    }),
    ({
        'expected': {
            'stat': 1
        },
        'input': {
            'cvr': {
                'ranks': [
                    [BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.OVERVOTE],
                    [BallotMarks.SKIPPED, BallotMarks.WRITEIN, 'B']
                ],
                'weight': [1, 1]
            }
        }
    }),
    ({
        'expected': {
            'stat': 5
        },
        'input': {
            'cvr': {
                'ranks': [
                    [BallotMarks.WRITEIN, BallotMarks.OVERVOTE, BallotMarks.OVERVOTE],
                    [BallotMarks.WRITEIN, BallotMarks.SKIPPED, 'B']
                ],
                'weight': [1, 5]
            }
        }
    }),
]


@pytest.mark.parametrize("param", params)
def test_ranked_multiple(param):
    cast_vote_record = CastVoteRecord(parsed_cvr=param['input']['cvr'])
    computed_stat = cast_vote_record.stats()['ranked_multiple'].item()
    assert param['expected']['stat'] == computed_stat


params = [
    ({
        'expected': {
            'stat': [0]
        },
        'input': {
            'cvr': {
                'ranks': [[BallotMarks.SKIPPED, 'A', BallotMarks.OVERVOTE]],
                'weight': [1],
                'split': [1]
            }
        }
    }),
    ({
        'expected': {
            'stat': [0, 1]
        },
        'input': {
            'cvr': {
                'ranks': [
                    [BallotMarks.SKIPPED, 'A', BallotMarks.OVERVOTE],
                    ['A', 'B', 'C']
                    ],
                'weight': [1, 1],
                'split': [1, 2]
            }
        }
    }),
]


@pytest.mark.parametrize("param", params)
def test_split_ranked_multiple(param):
    cast_vote_record = CastVoteRecord(parsed_cvr=param['input']['cvr'], split_fields=['split'])
    computed_stat = cast_vote_record.stats(add_split_stats=True)['split_ranked_multiple'].tolist()
    assert param['expected']['stat'] == computed_stat


params = [
    ({
        'expected': {
            'stat': 0
        },
        'input': {
            'cvr': {
                'ranks': [[BallotMarks.SKIPPED, 'A', BallotMarks.OVERVOTE]],
                'weight': [1]
            }
        }
    }),
    ({
        'expected': {
            'stat': 1
        },
        'input': {
            'cvr': {
                'ranks': [[BallotMarks.WRITEIN, 'A', 'B']],
                'weight': [1]
            }
        }
    }),
    ({
        'expected': {
            'stat': 1
        },
        'input': {
            'cvr': {
                'ranks': [
                    [BallotMarks.WRITEIN, 'A', 'B'],
                    [BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.SKIPPED]
                ],
                'weight': [1, 1]
            }
        }
    }),
    ({
        'expected': {
            'stat': 2
        },
        'input': {
            'cvr': {
                'ranks': [
                    [BallotMarks.WRITEIN, 'A', 'B'],
                    ['A', 'B', 'C']
                ],
                'weight': [1, 1]
            }
        }
    }),
]


@pytest.mark.parametrize("param", params)
def test_ranked_3_or_more(param):
    cast_vote_record = CastVoteRecord(parsed_cvr=param['input']['cvr'])
    computed_stat = cast_vote_record.stats()['ranked_3_or_more'].item()
    assert param['expected']['stat'] == computed_stat


params = [
    ({
        'expected': {
            'stat': [0]
        },
        'input': {
            'cvr': {
                'ranks': [[BallotMarks.SKIPPED, 'A', BallotMarks.OVERVOTE]],
                'weight': [1],
                'split': [1]
            }
        }
    }),
    ({
        'expected': {
            'stat': [0, 1]
        },
        'input': {
            'cvr': {
                'ranks': [
                    [BallotMarks.SKIPPED, 'A', BallotMarks.OVERVOTE],
                    ['A', 'B', 'C']
                    ],
                'weight': [1, 1],
                'split': [1, 2]
            }
        }
    }),
]


@pytest.mark.parametrize("param", params)
def test_split_ranked_3_or_more(param):
    cast_vote_record = CastVoteRecord(parsed_cvr=param['input']['cvr'], split_fields=['split'])
    computed_stat = cast_vote_record.stats(add_split_stats=True)['split_ranked_3_or_more'].tolist()
    assert param['expected']['stat'] == computed_stat


params = [
    ({
        'expected': {
            'stat': 1
        },
        'input': {
            'cvr': {
                'ranks': [[BallotMarks.SKIPPED, 'A', BallotMarks.OVERVOTE]],
                'weight': [1]
            }
        }
    }),
    ({
        'expected': {
            'stat': 0
        },
        'input': {
            'cvr': {
                'ranks': [
                    [BallotMarks.SKIPPED, 'A', BallotMarks.OVERVOTE],
                    [BallotMarks.SKIPPED, 'B', BallotMarks.OVERVOTE],
                ],
                'weight': [1, 1]
            }
        }
    }),
    ({
        'expected': {
            'stat': 1
        },
        'input': {
            'cvr': {
                'ranks': [
                    [BallotMarks.SKIPPED, 'A', 'B'],
                    [BallotMarks.SKIPPED, 'B', BallotMarks.OVERVOTE],
                ],
                'weight': [1, 1]
            }
        }
    }),
    ({
        'expected': {
            'stat': 1
        },
        'input': {
            'cvr': {
                'ranks': [
                    [BallotMarks.SKIPPED, 'A', 'B'],
                    ['C', 'E', 'D'],
                ],
                'weight': [1, 1]
            }
        }
    }),
    ({
        'expected': {
            'stat': 0
        },
        'input': {
            'cvr': {
                'ranks': [
                    [BallotMarks.SKIPPED, 'A'],
                    ['C', BallotMarks.SKIPPED],
                    ['B', BallotMarks.OVERVOTE],
                    ['D', BallotMarks.SKIPPED]
                ],
                'weight': [1, 1, 1, 1]
            }
        }
    }),
]


@pytest.mark.parametrize("param", params)
def test_total_fully_ranked(param):
    cast_vote_record = CastVoteRecord(parsed_cvr=param['input']['cvr'])
    computed_stat = cast_vote_record.stats()['total_fully_ranked'].item()
    assert param['expected']['stat'] == computed_stat


params = [
    ({
        'expected': {
            'stat': [1, 0]
        },
        'input': {
            'cvr': {
                'ranks': [
                    [BallotMarks.SKIPPED, 'A', BallotMarks.OVERVOTE],
                    [BallotMarks.SKIPPED, 'A', 'B'],
                    [BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.OVERVOTE],
                    [BallotMarks.SKIPPED, 'A', BallotMarks.OVERVOTE]
                ],
                'weight': [1, 1, 1, 1],
                'split': [1, 1, 2, 2]
            }
        }
    }),
    ({
        'expected': {
            'stat': [2, 0]
        },
        'input': {
            'cvr': {
                'ranks': [
                    ['B', 'A', 'C'],
                    ['D', 'A', 'B'],
                    [BallotMarks.SKIPPED, 'E', 'A'],
                    ['A', 'A', 'D']
                ],
                'weight': [1, 1, 1, 1],
                'split': [1, 1, 2, 2]
            }
        }
    }),
]


@pytest.mark.parametrize("param", params)
def test_split_total_fully_ranked(param):
    cast_vote_record = CastVoteRecord(parsed_cvr=param['input']['cvr'], split_fields=['split'])
    computed_stat = cast_vote_record.stats(add_split_stats=True)['split_total_fully_ranked'].tolist()
    assert param['expected']['stat'] == computed_stat


params = [
    ({
        'expected': {
            'stat': 0
        },
        'input': {
            'cvr': {
                'ranks': [[BallotMarks.SKIPPED, 'A', BallotMarks.OVERVOTE]],
                'weight': [1]
            }
        }
    }),
    ({
        'expected': {
            'stat': 1
        },
        'input': {
            'cvr': {
                'ranks': [[BallotMarks.SKIPPED, 'A', 'A']],
                'weight': [1]
            }
        }
    }),
    ({
        'expected': {
            'stat': 1
        },
        'input': {
            'cvr': {
                'ranks': [
                    [BallotMarks.SKIPPED, BallotMarks.WRITEIN, BallotMarks.WRITEIN],
                    [BallotMarks.SKIPPED, BallotMarks.WRITEIN, 'write-in']
                    ],
                'weight': [1, 1]
            }
        }
    }),
]


@pytest.mark.parametrize("param", params)
def test_includes_duplicate_ranking(param):
    cast_vote_record = CastVoteRecord(parsed_cvr=param['input']['cvr'])
    computed_stat = cast_vote_record.stats()['includes_duplicate_ranking'].item()
    assert param['expected']['stat'] == computed_stat


params = [
    ({
        'expected': {
            'stat': [0, 0]
        },
        'input': {
            'cvr': {
                'ranks': [
                    [BallotMarks.SKIPPED, 'A', BallotMarks.OVERVOTE],
                    [BallotMarks.SKIPPED, 'A', 'B'],
                    [BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.OVERVOTE],
                    [BallotMarks.SKIPPED, 'A', BallotMarks.OVERVOTE]
                ],
                'weight': [1, 1, 1, 1],
                'split': [1, 1, 2, 2]
            }
        }
    }),
    ({
        'expected': {
            'stat': [0, 1]
        },
        'input': {
            'cvr': {
                'ranks': [
                    ['B', 'A', 'C'],
                    ['D', 'A', 'B'],
                    [BallotMarks.SKIPPED, 'E', 'A'],
                    ['A', 'A', 'D']
                ],
                'weight': [1, 1, 1, 1],
                'split': [1, 1, 2, 2]
            }
        }
    }),
]


@pytest.mark.parametrize("param", params)
def test_split_includes_duplicate_ranking(param):
    cast_vote_record = CastVoteRecord(parsed_cvr=param['input']['cvr'], split_fields=['split'])
    computed_stat = cast_vote_record.stats(add_split_stats=True)['split_includes_duplicate_ranking'].tolist()
    assert param['expected']['stat'] == computed_stat


params = [
    ({
        'expected': {
            'stat': 1
        },
        'input': {
            'cvr': {
                'ranks': [[BallotMarks.SKIPPED, 'A', BallotMarks.OVERVOTE]],
                'weight': [1]
            }
        }
    }),
    ({
        'expected': {
            'stat': 1
        },
        'input': {
            'cvr': {
                'ranks': [[BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.OVERVOTE]],
                'weight': [1]
            }
        }
    }),
    ({
        'expected': {
            'stat': 1
        },
        'input': {
            'cvr': {
                'ranks': [
                    [BallotMarks.SKIPPED, BallotMarks.SKIPPED, 'A'],
                    ['A', BallotMarks.SKIPPED, BallotMarks.SKIPPED]
                    ],
                'weight': [1, 1]
            }
        }
    }),
]


@pytest.mark.parametrize("param", params)
def test_includes_skipped_ranking(param):
    cast_vote_record = CastVoteRecord(parsed_cvr=param['input']['cvr'])
    computed_stat = cast_vote_record.stats()['includes_skipped_ranking'].item()
    assert param['expected']['stat'] == computed_stat


params = [
    ({
        'expected': {
            'stat': [2, 2]
        },
        'input': {
            'cvr': {
                'ranks': [
                    [BallotMarks.SKIPPED, 'A', BallotMarks.OVERVOTE],
                    [BallotMarks.SKIPPED, 'A', 'B'],
                    [BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.OVERVOTE],
                    [BallotMarks.SKIPPED, 'A', BallotMarks.OVERVOTE]
                ],
                'weight': [1, 1, 1, 1],
                'split': [1, 1, 2, 2]
            }
        }
    }),
    ({
        'expected': {
            'stat': [0, 1]
        },
        'input': {
            'cvr': {
                'ranks': [
                    ['B', 'A', 'C'],
                    ['D', 'A', 'B'],
                    [BallotMarks.SKIPPED, 'E', 'A'],
                    ['A', 'A', 'D']
                ],
                'weight': [1, 1, 1, 1],
                'split': [1, 1, 2, 2]
            }
        }
    }),
]


@pytest.mark.parametrize("param", params)
def test_split_includes_skipped_ranking(param):
    cast_vote_record = CastVoteRecord(parsed_cvr=param['input']['cvr'], split_fields=['split'])
    computed_stat = cast_vote_record.stats(add_split_stats=True)['split_includes_skipped_ranking'].tolist()
    assert param['expected']['stat'] == computed_stat


params = [
    ({
        'expected': {
            'stat': 1
        },
        'input': {
            'cvr': {
                'ranks': [[BallotMarks.SKIPPED, 'A', BallotMarks.OVERVOTE]],
                'weight': [1]
            }
        }
    }),
    ({
        'expected': {
            'stat': 0
        },
        'input': {
            'cvr': {
                'ranks': [[BallotMarks.SKIPPED, 'A', BallotMarks.WRITEIN]],
                'weight': [1]
            }
        }
    }),
]


@pytest.mark.parametrize("param", params)
def test_includes_overvote_ranking(param):
    cast_vote_record = CastVoteRecord(parsed_cvr=param['input']['cvr'])
    computed_stat = cast_vote_record.stats()['includes_overvote_ranking'].item()
    assert param['expected']['stat'] == computed_stat


params = [
    ({
        'expected': {
            'stat': [1, 2]
        },
        'input': {
            'cvr': {
                'ranks': [
                    [BallotMarks.SKIPPED, 'A', BallotMarks.OVERVOTE],
                    [BallotMarks.SKIPPED, 'A', 'B'],
                    [BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.OVERVOTE],
                    [BallotMarks.SKIPPED, 'A', BallotMarks.OVERVOTE]
                ],
                'weight': [1, 1, 1, 1],
                'split': [1, 1, 2, 2]
            }
        }
    }),
    ({
        'expected': {
            'stat': [0, 0]
        },
        'input': {
            'cvr': {
                'ranks': [
                    ['B', 'A', 'C'],
                    ['D', 'A', 'B'],
                    [BallotMarks.SKIPPED, 'E', 'A'],
                    ['A', 'A', 'D']
                ],
                'weight': [1, 1, 1, 1],
                'split': [1, 1, 2, 2]
            }
        }
    }),
]


@pytest.mark.parametrize("param", params)
def test_split_includes_overvote_ranking(param):
    cast_vote_record = CastVoteRecord(parsed_cvr=param['input']['cvr'], split_fields=['split'])
    computed_stat = cast_vote_record.stats(add_split_stats=True)['split_includes_overvote_ranking'].tolist()
    assert param['expected']['stat'] == computed_stat


params = [
    ({
        'expected': {
            'stat': 1
        },
        'input': {
            'cvr': {
                'ranks': [[BallotMarks.SKIPPED, 'A', 'A', BallotMarks.OVERVOTE]],
                'weight': [1]
            }
        }
    }),
    ({
        'expected': {
            'stat': 4
        },
        'input': {
            'cvr': {
                'ranks': [
                    [BallotMarks.SKIPPED, 'A', 'B', 'C'],
                    ['C', 'A', 'B', BallotMarks.OVERVOTE],
                    ['C', 'A', 'A', BallotMarks.SKIPPED],
                    [BallotMarks.SKIPPED, 'A', 'A', BallotMarks.OVERVOTE]
                    ],
                'weight': [1, 1, 1, 1]
            }
        }
    }),
    ({
        'expected': {
            'stat': 0
        },
        'input': {
            'cvr': {
                'ranks': [
                    ['D', 'A', 'B', 'C'],
                    ['C', 'A', 'B', BallotMarks.WRITEIN],
                    ['C', 'A', 'D', BallotMarks.WRITEIN],
                    [BallotMarks.WRITEIN, 'B', 'A', BallotMarks.SKIPPED]
                    ],
                'weight': [1, 1, 1, 1]
            }
        }
    }),
]


@pytest.mark.parametrize("param", params)
def test_total_irregular(param):
    cast_vote_record = CastVoteRecord(parsed_cvr=param['input']['cvr'])
    computed_stat = cast_vote_record.stats()['total_irregular'].item()
    assert param['expected']['stat'] == computed_stat


params = [
    ({
        'expected': {
            'stat': [2, 2]
        },
        'input': {
            'cvr': {
                'ranks': [
                    [BallotMarks.SKIPPED, 'A', BallotMarks.OVERVOTE],
                    [BallotMarks.SKIPPED, 'A', 'B'],
                    [BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.OVERVOTE],
                    [BallotMarks.SKIPPED, 'A', BallotMarks.OVERVOTE]
                ],
                'weight': [1, 1, 1, 1],
                'split': [1, 1, 2, 2]
            }
        }
    }),
    ({
        'expected': {
            'stat': [0, 2]
        },
        'input': {
            'cvr': {
                'ranks': [
                    ['B', 'A', 'C'],
                    ['D', 'A', 'B'],
                    [BallotMarks.SKIPPED, 'E', 'A'],
                    ['A', 'A', 'D']
                ],
                'weight': [1, 1, 1, 1],
                'split': [1, 1, 2, 2]
            }
        }
    }),
]


@pytest.mark.parametrize("param", params)
def test_split_total_irregular(param):
    cast_vote_record = CastVoteRecord(parsed_cvr=param['input']['cvr'], split_fields=['split'])
    computed_stat = cast_vote_record.stats(add_split_stats=True)['split_total_irregular'].tolist()
    assert param['expected']['stat'] == computed_stat


params = [
    ({
        'expected': {
            'stat': 1
        },
        'input': {
            'cvr': {
                'ranks': [[BallotMarks.SKIPPED, 'A', 'A', BallotMarks.OVERVOTE]],
                'weight': [1]
            }
        }
    }),
    ({
        'expected': {
            'stat': 2
        },
        'input': {
            'cvr': {
                'ranks': [['A'], ['B']],
                'weight': [1, 1]
            }
        }
    }),
    ({
        'expected': {
            'stat': 5
        },
        'input': {
            'cvr': {
                'ranks': [['A'], ['B']],
                'weight': [2.5, 2.5]
            }
        }
    }),
]


@pytest.mark.parametrize("param", params)
def test_total_ballots(param):
    cast_vote_record = CastVoteRecord(parsed_cvr=param['input']['cvr'])
    computed_stat = cast_vote_record.stats()['total_ballots'].item()
    assert param['expected']['stat'] == computed_stat


params = [
    ({
        'expected': {
            'stat': [2, 2]
        },
        'input': {
            'cvr': {
                'ranks': [
                    [BallotMarks.SKIPPED, 'A', BallotMarks.OVERVOTE],
                    [BallotMarks.SKIPPED, 'A', 'B'],
                    [BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.OVERVOTE],
                    [BallotMarks.SKIPPED, 'A', BallotMarks.OVERVOTE]
                ],
                'weight': [1, 1, 1, 1],
                'split': [1, 1, 2, 2]
            }
        }
    }),
    ({
        'expected': {
            'stat': [2, 2]
        },
        'input': {
            'cvr': {
                'ranks': [
                    ['B', 'A', 'C'],
                    ['D', 'A', 'B'],
                    [BallotMarks.SKIPPED, 'E', 'A'],
                    ['A', 'A', 'D']
                ],
                'weight': [1, 1, 1, 1],
                'split': [1, 1, 2, 2]
            }
        }
    }),
]


@pytest.mark.parametrize("param", params)
def test_split_total_ballots(param):
    cast_vote_record = CastVoteRecord(parsed_cvr=param['input']['cvr'], split_fields=['split'])
    computed_stat = cast_vote_record.stats(add_split_stats=True)['split_total_ballots'].tolist()
    assert param['expected']['stat'] == computed_stat


params = [
    ({
        'expected': {
            'stat': 1
        },
        'input': {
            'cvr': {
                'ranks': [
                    [BallotMarks.SKIPPED, 'A', 'A', BallotMarks.OVERVOTE],
                    [BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.SKIPPED]
                    ],
                'weight': [1, 1]
            }
        }
    }),
]


@pytest.mark.parametrize("param", params)
def test_total_undervote(param):
    cast_vote_record = CastVoteRecord(parsed_cvr=param['input']['cvr'])
    computed_stat = cast_vote_record.stats()['total_undervote'].item()
    assert param['expected']['stat'] == computed_stat


params = [
    ({
        'expected': {
            'stat': [0, 1]
        },
        'input': {
            'cvr': {
                'ranks': [
                    [BallotMarks.SKIPPED, 'A', BallotMarks.OVERVOTE],
                    [BallotMarks.SKIPPED, 'A', 'B'],
                    [BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.OVERVOTE],
                    [BallotMarks.SKIPPED, 'A', BallotMarks.OVERVOTE],
                    [BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.SKIPPED]
                ],
                'weight': [1, 1, 1, 1, 1],
                'split': [1, 1, 2, 2, 2]
            }
        }
    }),
    ({
        'expected': {
            'stat': [0, 0]
        },
        'input': {
            'cvr': {
                'ranks': [
                    ['B', 'A', 'C'],
                    ['D', 'A', 'B'],
                    [BallotMarks.SKIPPED, 'E', 'A'],
                    ['A', 'A', 'D']
                ],
                'weight': [1, 1, 1, 1],
                'split': [1, 1, 2, 2]
            }
        }
    }),
]


@pytest.mark.parametrize("param", params)
def test_split_total_undervote(param):
    cast_vote_record = CastVoteRecord(parsed_cvr=param['input']['cvr'], split_fields=['split'])
    computed_stat = cast_vote_record.stats(add_split_stats=True)['split_total_undervote'].tolist()
    assert param['expected']['stat'] == computed_stat


params = [
    ({
        'expected': {
            'stat': 1
        },
        'input': {
            'cvr': {
                'ranks': [
                    [BallotMarks.SKIPPED, 'A', 'A', BallotMarks.OVERVOTE],
                    [BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.SKIPPED]
                    ],
                'weight': [1, 1]
            }
        }
    }),
    ({
        'expected': {
            'stat': 1.5
        },
        'input': {
            'cvr': {
                'ranks': [
                    [BallotMarks.SKIPPED, 'A', 'A', BallotMarks.OVERVOTE],
                    [BallotMarks.SKIPPED, 'A', 'B', BallotMarks.SKIPPED]
                    ],
                'weight': [1, 1]
            }
        }
    }),
    ({
        'expected': {
            'stat': round(11/6, 3)
        },
        'input': {
            'cvr': {
                'ranks': [
                    [BallotMarks.SKIPPED, 'A', 'A', BallotMarks.OVERVOTE],
                    [BallotMarks.SKIPPED, 'A', 'B', BallotMarks.SKIPPED]
                    ],
                'weight': [1, 5]
            }
        }
    }),
]


@pytest.mark.parametrize("param", params)
def test_mean_rankings_used(param):
    cast_vote_record = CastVoteRecord(parsed_cvr=param['input']['cvr'])
    computed_stat = cast_vote_record.stats()['mean_rankings_used'].item()
    assert param['expected']['stat'] == computed_stat


params = [
    ({
        'expected': {
            'stat': [1.5, 0.5]
        },
        'input': {
            'cvr': {
                'ranks': [
                    [BallotMarks.SKIPPED, 'A', BallotMarks.OVERVOTE],
                    [BallotMarks.SKIPPED, 'A', 'B'],
                    [BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.OVERVOTE],
                    [BallotMarks.SKIPPED, 'A', BallotMarks.OVERVOTE],
                    [BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.SKIPPED]
                ],
                'weight': [1, 1, 1, 1, 1],
                'split': [1, 1, 2, 2, 2]
            }
        }
    }),
    ({
        'expected': {
            'stat': [3, 2]
        },
        'input': {
            'cvr': {
                'ranks': [
                    ['B', 'A', 'C'],
                    ['D', 'A', 'B'],
                    [BallotMarks.SKIPPED, 'E', 'A'],
                    ['A', 'A', 'D']
                ],
                'weight': [1, 1, 1, 1],
                'split': [1, 1, 2, 2]
            }
        }
    }),
]


@pytest.mark.parametrize("param", params)
def test_split_mean_rankings_used(param):
    cast_vote_record = CastVoteRecord(parsed_cvr=param['input']['cvr'], split_fields=['split'])
    computed_stat = cast_vote_record.stats(add_split_stats=True)['split_mean_rankings_used'].tolist()
    assert param['expected']['stat'] == computed_stat


params = [
    ({
        'expected': {
            'stat': 1
        },
        'input': {
            'cvr': {
                'ranks': [
                    [BallotMarks.SKIPPED, 'A', 'A', BallotMarks.OVERVOTE],
                    [BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.SKIPPED]
                    ],
                'weight': [1, 1]
            }
        }
    }),
    ({
        'expected': {
            'stat': 1.5
        },
        'input': {
            'cvr': {
                'ranks': [
                    [BallotMarks.SKIPPED, 'A', 'A', BallotMarks.OVERVOTE],
                    [BallotMarks.SKIPPED, 'A', 'B', BallotMarks.SKIPPED]
                    ],
                'weight': [1, 1]
            }
        }
    }),
    ({
        'expected': {
            'stat': 2
        },
        'input': {
            'cvr': {
                'ranks': [
                    [BallotMarks.SKIPPED, 'A', 'A', BallotMarks.OVERVOTE],
                    [BallotMarks.SKIPPED, 'A', 'B', BallotMarks.SKIPPED]
                    ],
                'weight': [1, 5]
            }
        }
    }),
]


@pytest.mark.parametrize("param", params)
def test_median_rankings_used(param):
    cast_vote_record = CastVoteRecord(parsed_cvr=param['input']['cvr'])
    computed_stat = cast_vote_record.stats()['median_rankings_used'].item()
    assert param['expected']['stat'] == computed_stat


params = [
    ({
        'expected': {
            'stat': [1.5, 0.5]
        },
        'input': {
            'cvr': {
                'ranks': [
                    [BallotMarks.SKIPPED, 'A', BallotMarks.OVERVOTE],
                    [BallotMarks.SKIPPED, 'A', 'B'],
                    [BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.OVERVOTE],
                    [BallotMarks.SKIPPED, 'A', BallotMarks.OVERVOTE],
                    [BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.SKIPPED]
                ],
                'weight': [1, 1, 1, 1, 1],
                'split': [1, 1, 2, 2, 2]
            }
        }
    }),
    ({
        'expected': {
            'stat': [3, 2]
        },
        'input': {
            'cvr': {
                'ranks': [
                    ['B', 'A', 'C'],
                    ['D', 'A', 'B'],
                    [BallotMarks.SKIPPED, 'E', 'A'],
                    ['A', 'A', 'D']
                ],
                'weight': [1, 1, 1, 6],
                'split': [1, 1, 2, 2]
            }
        }
    }),
]


@pytest.mark.parametrize("param", params)
def test_split_median_rankings_used(param):
    cast_vote_record = CastVoteRecord(parsed_cvr=param['input']['cvr'], split_fields=['split'])
    computed_stat = cast_vote_record.stats(add_split_stats=True)['split_median_rankings_used'].tolist()
    assert param['expected']['stat'] == computed_stat



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

import io
import pathlib
from glob import glob
from os.path import basename
from os.path import dirname
from os.path import join
from os.path import splitext

from setuptools import find_packages
from setuptools import setup


def read(*names, **kwargs):
    with io.open(
        join(dirname(__file__), *names),
        encoding=kwargs.get('encoding', 'utf8')
    ) as fh:
        return fh.read()

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()


setup(
    name='rcv_cruncher',
    version='0.0.6',
    description='Analyze RCV elections',
    # long_description='%s\n%s' % (
    #     re.compile('^.. start-badges.*^.. end-badges', re.M | re.S).sub('', read('README.rst')),
    #     re.sub(':[a-z]+:`~?(.*?)`', r'``\1``', read('CHANGELOG.rst'))
    # ),
    long_description=README,
    long_description_content_type="text/markdown",
    #long_description="",
    author='Chris Zawora',
    author_email='christopher.zawora@gmail.com',
    url='https://github.com/fairvotereform/rcv_cruncher',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    py_modules=[splitext(basename(path))[0] for path in glob('src/*.py')],
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        # complete classifier list: http://pypi.python.org/pypi?%3Aaction=list_classifiers
        # 'Development Status :: 5 - Production/Stable',
        # 'Intended Audience :: Developers',
        'Operating System :: Unix',
        'Operating System :: POSIX',
        'Operating System :: Microsoft :: Windows',
        # 'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        # 'Programming Language :: Python :: 3 :: Only',
        # 'Programming Language :: Python :: 3.6',
        # 'Programming Language :: Python :: 3.7',
        # 'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        # uncomment if you test on these interpreters:
        # 'Programming Language :: Python :: Implementation :: IronPython',
        # 'Programming Language :: Python :: Implementation :: Jython',
        # 'Programming Language :: Python :: Implementation :: Stackless',
        'Topic :: Utilities',
    ],
    project_urls={
        'Documentation': 'https://rcv_cruncher.readthedocs.io/',
        'Changelog': 'https://rcv_cruncher.readthedocs.io/en/latest/changelog.html',
        'Issue Tracker': 'https://github.com/fairvotereform/rcv_cruncher/issues',
    },
    keywords=[
        # eg: 'keyword1', 'keyword2', 'keyword3',
    ],
    python_requires='>=3.6',
    install_requires=[
        'tqdm>=4.56.0',
        'pandas>=1.2.0',
        'xmltodict>=0.12.0',
        'weightedstats>=0.4.1'
    ],
    extras_require={},
    entry_points={
        'console_scripts': [
            'rcv-cruncher = rcv_cruncher.cli:main',
        ]
    },
)




import pytest

from rcv_cruncher.marks import BallotMarks
from rcv_cruncher.rcv.variants import SingleWinner

# testing:

# tabulations
# first_round_winner_vote
# final_round_winner_vote
# first_round_winner_percent
# final_round_winner_percent
# first_round_winner_place
# condorcet
# come_from_behind
# ranked_winner
# final_round_winner_votes_over_first_round_valid
# win_threshold
# number_of_winners
# winner
# number_of_rounds
# winners_consensus_value
# first_round_active_votes
# final_round_active_votes
# total_pretally_exhausted
# total_posttally_exhausted
# total_posttally_exhausted_by_overvote
# total_posttally_exhausted_by_skipped_rankings
# total_posttally_exhausted_by_abstention
# total_posttally_exhausted_by_rank_limit
# total_posttally_exhausted_by_duplicate_rankings


params = [
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'B', 'C', 'D'],
                    ['A', BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.WRITEIN],
                    ['write-in', 'A', 'C', BallotMarks.OVERVOTE],
                    ['write-in', 'B', 'B', BallotMarks.OVERVOTE],
                    ['C', 'A', 'B', 'B']
                ]
            }
        },
        'expected': {
            'n_tabulation': 1,
            'n_round': 2,
            'rounds': [
                {
                    'A': 2,
                    'B': 0,
                    'C': 1,
                    'D': 0,
                    BallotMarks.WRITEIN: 2

                },
                {
                    'A': 3,
                    'B': 0,
                    'C': 0,
                    'D': 0,
                    BallotMarks.WRITEIN: 2
                }
            ]
        }
    }),
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'B', 'C', 'D'],
                    ['A', BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.WRITEIN],
                    ['write-in', 'A', 'C', BallotMarks.OVERVOTE],
                    ['write-in', 'B', 'B', BallotMarks.OVERVOTE],
                    ['C', 'D', BallotMarks.WRITEIN, 'B']
                ]
            }
        },
        'expected': {
            'n_tabulation': 1,
            'n_round': 2,
            'rounds': [
                {
                    'A': 2,
                    'B': 0,
                    'C': 1,
                    'D': 0,
                    BallotMarks.WRITEIN: 2

                },
                {
                    'A': 2,
                    'B': 0,
                    'C': 0,
                    'D': 0,
                    BallotMarks.WRITEIN: 3
                }
            ]
        }
    }),
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'B', 'C', 'D'],
                    ['A', BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.WRITEIN],
                    ['A', BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.SKIPPED],
                    ['write-in', 'A', 'C', BallotMarks.OVERVOTE],
                    ['write-in', 'B', 'B', BallotMarks.OVERVOTE],
                    ['C', 'D', BallotMarks.WRITEIN, 'B'],
                    ['C', 'D', 'D', 'B'],
                    ['D', 'B', BallotMarks.WRITEIN, 'D']
                ]
            }
        },
        'expected': {
            'n_tabulation': 1,
            'n_round': 3,
            'rounds': [
                {
                    'A': 3,
                    'B': 0,
                    'C': 2,
                    'D': 1,
                    BallotMarks.WRITEIN: 2

                },
                {
                    'A': 3,
                    'B': 0,
                    'C': 2,
                    'D': 0,
                    BallotMarks.WRITEIN: 3
                },
                {
                    'A': 3,
                    'B': 0,
                    'C': 0,
                    'D': 0,
                    BallotMarks.WRITEIN: 4
                }
            ]
        }
    }),
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'B', 'C', 'D'],
                    ['A', BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.WRITEIN],
                    ['A', BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.SKIPPED],
                    ['write-in', 'A', 'C', BallotMarks.OVERVOTE],
                    ['write-in', 'B', 'B', BallotMarks.OVERVOTE],
                    ['C', 'D', BallotMarks.WRITEIN, 'B'],
                    ['C', 'D', 'D', 'B'],
                    ['D', 'B', BallotMarks.WRITEIN, 'D']
                ],
                'weight': [2, 2, 2, 2, 1, 1, 1, 1]
            }
        },
        'expected': {
            'n_tabulation': 1,
            'n_round': 3,
            'rounds': [
                {
                    'A': 6,
                    'B': 0,
                    'C': 2,
                    'D': 1,
                    BallotMarks.WRITEIN: 3

                },
                {
                    'A': 6,
                    'B': 0,
                    'C': 2,
                    'D': 0,
                    BallotMarks.WRITEIN: 4
                },
                {
                    'A': 6,
                    'B': 0,
                    'C': 0,
                    'D': 0,
                    BallotMarks.WRITEIN: 5
                }
            ]
        }
    }),
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'B', 'C', 'D'],
                    ['A', BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.WRITEIN],
                    ['A', BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.SKIPPED],
                    ['write-in', 'A', 'C', BallotMarks.OVERVOTE],
                    ['write-in', 'B', 'B', BallotMarks.OVERVOTE],
                    ['C', 'D', BallotMarks.WRITEIN, 'B'],
                    ['C', 'D', 'D', 'B'],
                    ['D', 'B', BallotMarks.WRITEIN, 'D']
                ],
                'weight': [2, 2, 2, 2, 1, 1, 1, 1]
            },

        },
        'expected': {
            'n_tabulation': 1,
            'n_round': 3,
            'rounds': [
                {
                    'A': 6,
                    'B': 0,
                    'C': 2,
                    'D': 1,
                    BallotMarks.WRITEIN: 3

                },
                {
                    'A': 6,
                    'B': 0,
                    'C': 2,
                    'D': 0,
                    BallotMarks.WRITEIN: 4
                },
                {
                    'A': 6,
                    'B': 0,
                    'C': 0,
                    'D': 0,
                    BallotMarks.WRITEIN: 5
                }
            ]
        }
    }),
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'C', 'C', 'D'],
                    [BallotMarks.OVERVOTE, 'A', 'B', 'C'],
                    ['B', BallotMarks.OVERVOTE, 'A', 'B'],
                    ['B', 'A', 'A', 'C'],
                    ['C', BallotMarks.SKIPPED, BallotMarks.SKIPPED, 'A'],
                    ['C', 'A', 'B', 'E'],
                    ['D', BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.SKIPPED],
                    ['E', BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.SKIPPED]
                ],
                'weight': [1, 1, 1, 1, 1, 1, 5, 5],
                'precinct': [1, 1, 1, 2, 2, 2, 3, 3]
            },
            'split_fields': ['precinct'],
            'exhaust_on_duplicate_candidate_marks': True,
            'exhaust_on_overvote_marks': True,
            'exhaust_on_repeated_skipped_marks': True
        },
        'expected': {
            'n_tabulation': 1,
            'n_round': 4,
            'rounds': [
                {
                    'A': 1,
                    'B': 2,
                    'C': 2,
                    'D': 5,
                    'E': 5
                },
                {
                    'A': 0,
                    'B': 2,
                    'C': 3,
                    'D': 5,
                    'E': 5
                },
                {
                    'A': 0,
                    'B': 0,
                    'C': 3,
                    'D': 5,
                    'E': 5
                },
                {
                    'A': 0,
                    'B': 0,
                    'C': 0,
                    'D': 5,
                    'E': 6
                }
            ]
        }
    })
]

# 'exhaust_on_duplicate_candidate_marks': False,
# 'exhaust_on_overvote_marks': False,
# 'exhaust_on_repeated_skipped_marks': False,
# 'treat_combined_writeins_as_exhaustable_duplicates': True,
# 'combine_writein_marks': True,
# 'exclude_writein_marks': False


@pytest.mark.parametrize("param", params)
def test_tabulation(param):
    rcv = SingleWinner(**param['input'])

    # confirm tabulation num
    n_tabulations = rcv.n_tabulations()
    assert n_tabulations == param['expected']['n_tabulation']

    # confirm round num
    n_round = rcv.n_rounds(tabulation_num=1)
    assert n_round == param['expected']['n_round']

    # confirm round tallies
    tally_dict = [rcv.get_round_tally_dict(round_num=i) for i in range(1, n_round + 1)]
    tally_dict = [{k: float(v) for k, v in d.items()} for d in tally_dict]
    assert tally_dict == param['expected']['rounds']


params = [
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'B', 'C', 'D'],
                    ['A', BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.WRITEIN],
                    ['write-in', 'A', 'C', BallotMarks.OVERVOTE],
                    ['write-in', 'B', 'B', BallotMarks.OVERVOTE],
                    ['C', 'A', 'B', 'B']
                ]
            }
        },
        'expected': {
            'stat': 2
        }
    }),
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'B', 'C', 'D'],
                    ['A', BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.WRITEIN],
                    ['write-in', 'A', 'C', BallotMarks.OVERVOTE],
                    ['write-in', 'B', 'B', BallotMarks.OVERVOTE],
                    ['C', 'D', BallotMarks.WRITEIN, 'B']
                ]
            }
        },
        'expected': {
            'stat': 2
        }
    }),
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'B', 'C', 'D'],
                    ['A', BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.WRITEIN],
                    ['A', BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.SKIPPED],
                    ['write-in', 'A', 'C', BallotMarks.OVERVOTE],
                    ['write-in', 'B', 'B', BallotMarks.OVERVOTE],
                    ['C', 'D', BallotMarks.WRITEIN, 'B'],
                    ['C', 'D', 'D', 'B'],
                    ['D', 'B', BallotMarks.WRITEIN, 'D']
                ],
                'precinct': [1, 1, 1, 2, 2, 2, 3, 3]
            },
            'split_fields': ['precinct']
        },
        'expected': {
            'stat': 2
        }
    })
]


@pytest.mark.parametrize("param", params)
def test_first_round_winner_vote(param):
    rcv = SingleWinner(**param['input'])
    assert rcv.stats()[0]['first_round_winner_vote'].item() == param['expected']['stat']


params = [
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'B', 'C', 'D'],
                    ['A', BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.WRITEIN],
                    ['write-in', 'A', 'C', BallotMarks.OVERVOTE],
                    ['write-in', 'B', 'B', BallotMarks.OVERVOTE],
                    ['C', 'A', 'B', 'B']
                ]
            }
        },
        'expected': {
            'stat': 3
        }
    }),
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'B', 'C', 'D'],
                    ['A', BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.WRITEIN],
                    ['write-in', 'A', 'C', BallotMarks.OVERVOTE],
                    ['write-in', 'B', 'B', BallotMarks.OVERVOTE],
                    ['C', 'D', BallotMarks.WRITEIN, 'B']
                ]
            }
        },
        'expected': {
            'stat': 3
        }
    }),
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'B', 'C', 'D'],
                    ['A', BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.WRITEIN],
                    ['A', BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.SKIPPED],
                    ['write-in', 'A', 'C', BallotMarks.OVERVOTE],
                    ['write-in', 'B', 'B', BallotMarks.OVERVOTE],
                    ['C', 'D', BallotMarks.WRITEIN, 'B'],
                    ['C', 'D', 'D', 'B'],
                    ['D', 'B', BallotMarks.WRITEIN, 'D']
                ],
                'precinct': [1, 1, 1, 2, 2, 2, 3, 3]
            },
            'split_fields': ['precinct']
        },
        'expected': {
            'stat': 4
        }
    })
]


@pytest.mark.parametrize("param", params)
def test_final_round_winner_vote(param):
    rcv = SingleWinner(**param['input'])
    assert rcv.stats()[0]['final_round_winner_vote'].item() == param['expected']['stat']


params = [
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'B', 'C', 'D'],
                    ['A', BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.WRITEIN],
                    ['write-in', 'A', 'C', BallotMarks.OVERVOTE],
                    ['write-in', 'B', 'B', BallotMarks.OVERVOTE],
                    ['C', 'A', 'B', 'B']
                ]
            }
        },
        'expected': {
            'stat': 40
        }
    }),
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'B', 'C', 'D'],
                    ['A', BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.WRITEIN],
                    ['write-in', 'A', 'C', BallotMarks.OVERVOTE],
                    ['write-in', 'B', 'B', BallotMarks.OVERVOTE],
                    ['C', 'D', BallotMarks.WRITEIN, 'B']
                ]
            }
        },
        'expected': {
            'stat': 40
        }
    }),
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'B', 'C', 'D'],
                    ['A', BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.WRITEIN],
                    ['A', BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.SKIPPED],
                    ['write-in', 'A', 'C', BallotMarks.OVERVOTE],
                    ['write-in', 'B', 'B', BallotMarks.OVERVOTE],
                    ['C', 'D', BallotMarks.WRITEIN, 'B'],
                    ['C', 'D', 'D', 'B'],
                    ['D', 'B', BallotMarks.WRITEIN, 'D']
                ],
                'precinct': [1, 1, 1, 2, 2, 2, 3, 3]
            },
            'split_fields': ['precinct']
        },
        'expected': {
            'stat': 25
        }
    })
]


@pytest.mark.parametrize("param", params)
def test_first_round_winner_percent(param):
    rcv = SingleWinner(**param['input'])
    assert rcv.stats()[0]['first_round_winner_percent'].item() == param['expected']['stat']


params = [
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'B', 'C', 'D'],
                    ['A', BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.WRITEIN],
                    ['write-in', 'A', 'C', BallotMarks.OVERVOTE],
                    ['write-in', 'B', 'B', BallotMarks.OVERVOTE],
                    ['C', 'A', 'B', 'B']
                ]
            }
        },
        'expected': {
            'stat': 60
        }
    }),
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'B', 'C', 'D'],
                    ['A', BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.WRITEIN],
                    ['write-in', 'A', 'C', BallotMarks.OVERVOTE],
                    ['write-in', 'B', 'B', BallotMarks.OVERVOTE],
                    ['C', 'D', BallotMarks.WRITEIN, 'B']
                ]
            }
        },
        'expected': {
            'stat': 60
        }
    }),
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'B', 'C', 'D'],
                    ['A', BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.WRITEIN],
                    ['A', BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.SKIPPED],
                    ['write-in', 'A', 'C', BallotMarks.OVERVOTE],
                    ['write-in', 'B', 'B', BallotMarks.OVERVOTE],
                    ['C', 'D', BallotMarks.WRITEIN, 'B'],
                    ['C', 'D', 'D', 'B'],
                    ['D', 'B', BallotMarks.WRITEIN, 'D']
                ],
                'precinct': [1, 1, 1, 2, 2, 2, 3, 3]
            },
            'split_fields': ['precinct']
        },
        'expected': {
            'stat': round(100 * 4/7, 3)
        }
    })
]


@pytest.mark.parametrize("param", params)
def test_final_round_winner_percent(param):
    rcv = SingleWinner(**param['input'])
    assert rcv.stats()[0]['final_round_winner_percent'].item() == param['expected']['stat']


params = [
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'B', 'C', 'D'],
                    ['A', BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.WRITEIN],
                    ['write-in', 'A', 'C', BallotMarks.OVERVOTE],
                    ['write-in', 'B', 'B', BallotMarks.OVERVOTE],
                    ['C', 'A', 'B', 'B']
                ]
            }
        },
        'expected': {
            'stat': 1
        }
    }),
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'B', 'C', 'D'],
                    ['A', BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.WRITEIN],
                    ['write-in', 'A', 'C', BallotMarks.OVERVOTE],
                    ['write-in', 'B', 'B', BallotMarks.OVERVOTE],
                    ['C', 'D', BallotMarks.WRITEIN, 'B']
                ]
            }
        },
        'expected': {
            'stat': 1
        }
    }),
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'B', 'C', 'D'],
                    ['A', BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.WRITEIN],
                    ['A', BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.SKIPPED],
                    ['write-in', 'A', 'C', BallotMarks.OVERVOTE],
                    ['write-in', 'B', 'B', BallotMarks.OVERVOTE],
                    ['C', 'D', BallotMarks.WRITEIN, 'B'],
                    ['C', 'D', 'D', 'B'],
                    ['D', 'B', BallotMarks.WRITEIN, 'D']
                ],
                'precinct': [1, 1, 1, 2, 2, 2, 3, 3]
            },
            'split_fields': ['precinct']
        },
        'expected': {
            'stat': 2
        }
    })
]


@pytest.mark.parametrize("param", params)
def test_first_round_winner_place(param):
    rcv = SingleWinner(**param['input'])
    assert rcv.stats()[0]['first_round_winner_place'].item() == param['expected']['stat']


params = [
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'B', 'C', 'D'],
                    ['A', BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.WRITEIN],
                    ['write-in', 'A', 'C', BallotMarks.OVERVOTE],
                    ['write-in', 'B', 'B', BallotMarks.OVERVOTE],
                    ['C', 'A', 'B', 'B']
                ]
            }
        },
        'expected': {
            'stat': True
        }
    }),
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'B', 'C', 'D'],
                    ['A', BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.WRITEIN],
                    ['write-in', 'A', 'C', BallotMarks.OVERVOTE],
                    ['write-in', 'B', 'B', BallotMarks.OVERVOTE],
                    ['C', 'D', BallotMarks.WRITEIN, 'B']
                ]
            }
        },
        'expected': {
            'stat': True
        }
    }),
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'B', 'C', 'D'],
                    ['A', BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.WRITEIN],
                    ['A', BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.SKIPPED],
                    ['write-in', 'A', 'C', BallotMarks.OVERVOTE],
                    ['write-in', 'B', 'B', BallotMarks.OVERVOTE],
                    ['C', 'D', BallotMarks.WRITEIN, 'B'],
                    ['C', 'D', 'D', 'B'],
                    ['D', 'B', BallotMarks.WRITEIN, 'D']
                ],
                'precinct': [1, 1, 1, 2, 2, 2, 3, 3]
            },
            'split_fields': ['precinct']
        },
        'expected': {
            'stat': False
        }
    })
]


@pytest.mark.parametrize("param", params)
def test_condorcet(param):
    rcv = SingleWinner(**param['input'])
    assert rcv.stats()[0]['condorcet'].item() == param['expected']['stat']


params = [
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'B', 'C', 'D'],
                    ['A', BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.WRITEIN],
                    ['write-in', 'A', 'C', BallotMarks.OVERVOTE],
                    ['write-in', 'B', 'B', BallotMarks.OVERVOTE],
                    ['C', 'A', 'B', 'B']
                ]
            }
        },
        'expected': {
            'stat': False
        }
    }),
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'B', 'C', 'D'],
                    ['A', BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.WRITEIN],
                    ['write-in', 'A', 'C', BallotMarks.OVERVOTE],
                    ['write-in', 'B', 'B', BallotMarks.OVERVOTE],
                    ['C', 'D', BallotMarks.WRITEIN, 'B']
                ]
            }
        },
        'expected': {
            'stat': False
        }
    }),
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'B', 'C', 'D'],
                    ['A', BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.WRITEIN],
                    ['A', BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.SKIPPED],
                    ['write-in', 'A', 'C', BallotMarks.OVERVOTE],
                    ['write-in', 'B', 'B', BallotMarks.OVERVOTE],
                    ['C', 'D', BallotMarks.WRITEIN, 'B'],
                    ['C', 'D', 'D', 'B'],
                    ['D', 'B', BallotMarks.WRITEIN, 'D']
                ],
                'precinct': [1, 1, 1, 2, 2, 2, 3, 3]
            },
            'split_fields': ['precinct']
        },
        'expected': {
            'stat': True
        }
    })
]


@pytest.mark.parametrize("param", params)
def test_come_from_behind(param):
    rcv = SingleWinner(**param['input'])
    assert rcv.stats()[0]['come_from_behind'].item() == param['expected']['stat']


params = [
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'B', 'C', 'D'],
                    ['A', BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.WRITEIN],
                    ['write-in', 'A', 'C', BallotMarks.OVERVOTE],
                    ['write-in', 'B', 'B', BallotMarks.OVERVOTE],
                    ['C', 'A', 'B', 'B']
                ]
            }
        },
        'expected': {
            'stat': 4
        }
    }),
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'B', 'C', 'D'],
                    ['A', BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.WRITEIN],
                    ['write-in', 'A', 'C', BallotMarks.OVERVOTE],
                    ['write-in', 'B', 'B', BallotMarks.OVERVOTE],
                    ['C', 'D', BallotMarks.WRITEIN, 'B']
                ]
            }
        },
        'expected': {
            'stat': 4
        }
    }),
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'B', 'C', 'D'],
                    ['A', BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.WRITEIN],
                    ['A', BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.SKIPPED],
                    ['write-in', 'A', 'C', BallotMarks.OVERVOTE],
                    ['write-in', 'B', 'B', BallotMarks.OVERVOTE],
                    ['C', 'D', BallotMarks.WRITEIN, 'B'],
                    ['C', 'D', 'D', 'B'],
                    ['D', 'B', BallotMarks.WRITEIN, 'D']
                ],
                'precinct': [1, 1, 1, 2, 2, 2, 3, 3]
            },
            'split_fields': ['precinct']
        },
        'expected': {
            'stat': 5
        }
    })
]


@pytest.mark.parametrize("param", params)
def test_ranked_winner(param):
    rcv = SingleWinner(**param['input'])
    assert rcv.stats()[0]['ranked_winner'].item() == param['expected']['stat']


params = [
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'B', 'C', 'D'],
                    ['A', BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.WRITEIN],
                    ['write-in', 'A', 'C', BallotMarks.OVERVOTE],
                    ['write-in', 'B', 'B', BallotMarks.OVERVOTE],
                    ['C', 'A', 'B', 'B']
                ]
            }
        },
        'expected': {
            'stat': 100 * 3/5
        }
    }),
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'B', 'C', 'D'],
                    ['A', BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.WRITEIN],
                    ['write-in', 'A', 'C', BallotMarks.OVERVOTE],
                    ['write-in', 'B', 'B', BallotMarks.OVERVOTE],
                    ['C', 'D', BallotMarks.WRITEIN, 'B']
                ]
            }
        },
        'expected': {
            'stat': 60
        }
    }),
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'B', 'C', 'D'],
                    ['A', BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.WRITEIN],
                    ['A', BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.SKIPPED],
                    ['write-in', 'A', 'C', BallotMarks.OVERVOTE],
                    ['write-in', 'B', 'B', BallotMarks.OVERVOTE],
                    ['C', 'D', BallotMarks.WRITEIN, 'B'],
                    ['C', 'D', 'D', 'B'],
                    ['D', 'B', BallotMarks.WRITEIN, 'D']
                ],
                'precinct': [1, 1, 1, 2, 2, 2, 3, 3]
            },
            'split_fields': ['precinct']
        },
        'expected': {
            'stat': 50
        }
    })
]


@pytest.mark.parametrize("param", params)
def test_final_round_winner_votes_over_first_round_active(param):
    rcv = SingleWinner(**param['input'])
    assert rcv.stats()[0]['final_round_winner_votes_over_first_round_active'].item() == param['expected']['stat']


params = [
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'B', 'C', 'D'],
                    ['A', BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.WRITEIN],
                    ['write-in', 'A', 'C', BallotMarks.OVERVOTE],
                    ['write-in', 'B', 'B', BallotMarks.OVERVOTE],
                    ['C', 'A', 'B', 'B']
                ]
            }
        },
        'expected': {
            'stat': None
        }
    }),
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'B', 'C', 'D'],
                    ['A', BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.WRITEIN],
                    ['write-in', 'A', 'C', BallotMarks.OVERVOTE],
                    ['write-in', 'B', 'B', BallotMarks.OVERVOTE],
                    ['C', 'D', BallotMarks.WRITEIN, 'B']
                ]
            }
        },
        'expected': {
            'stat': None
        }
    }),
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'B', 'C', 'D'],
                    ['A', BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.WRITEIN],
                    ['A', BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.SKIPPED],
                    ['write-in', 'A', 'C', BallotMarks.OVERVOTE],
                    ['write-in', 'B', 'B', BallotMarks.OVERVOTE],
                    ['C', 'D', BallotMarks.WRITEIN, 'B'],
                    ['C', 'D', 'D', 'B'],
                    ['D', 'B', BallotMarks.WRITEIN, 'D']
                ],
                'precinct': [1, 1, 1, 2, 2, 2, 3, 3]
            },
            'split_fields': ['precinct']
        },
        'expected': {
            'stat': None
        }
    })
]


@pytest.mark.parametrize("param", params)
def test_static_win_threshold(param):
    rcv = SingleWinner(**param['input'])
    assert rcv.stats()[0]['static_win_threshold'].item() == param['expected']['stat']


params = [
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'B', 'C', 'D'],
                    ['A', BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.WRITEIN],
                    ['write-in', 'A', 'C', BallotMarks.OVERVOTE],
                    ['write-in', 'B', 'B', BallotMarks.OVERVOTE],
                    ['C', 'A', 'B', 'B']
                ]
            }
        },
        'expected': {
            'stat': 1
        }
    }),
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'B', 'C', 'D'],
                    ['A', BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.WRITEIN],
                    ['write-in', 'A', 'C', BallotMarks.OVERVOTE],
                    ['write-in', 'B', 'B', BallotMarks.OVERVOTE],
                    ['C', 'D', BallotMarks.WRITEIN, 'B']
                ]
            }
        },
        'expected': {
            'stat': 1
        }
    }),
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'B', 'C', 'D'],
                    ['A', BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.WRITEIN],
                    ['A', BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.SKIPPED],
                    ['write-in', 'A', 'C', BallotMarks.OVERVOTE],
                    ['write-in', 'B', 'B', BallotMarks.OVERVOTE],
                    ['C', 'D', BallotMarks.WRITEIN, 'B'],
                    ['C', 'D', 'D', 'B'],
                    ['D', 'B', BallotMarks.WRITEIN, 'D']
                ],
                'precinct': [1, 1, 1, 2, 2, 2, 3, 3]
            },
            'split_fields': ['precinct']
        },
        'expected': {
            'stat': 1
        }
    })
]


@pytest.mark.parametrize("param", params)
def test_number_of_tabulation_winners(param):
    rcv = SingleWinner(**param['input'])
    assert rcv.stats()[0]['number_of_tabulation_winners'].item() == param['expected']['stat']


params = [
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'B', 'C', 'D'],
                    ['A', BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.WRITEIN],
                    ['write-in', 'A', 'C', BallotMarks.OVERVOTE],
                    ['write-in', 'B', 'B', BallotMarks.OVERVOTE],
                    ['C', 'A', 'B', 'B']
                ]
            }
        },
        'expected': {
            'stat': 'A'
        }
    }),
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'B', 'C', 'D'],
                    ['A', BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.WRITEIN],
                    ['write-in', 'A', 'C', BallotMarks.OVERVOTE],
                    ['write-in', 'B', 'B', BallotMarks.OVERVOTE],
                    ['C', 'D', BallotMarks.WRITEIN, 'B']
                ]
            }
        },
        'expected': {
            'stat': BallotMarks.WRITEIN
        }
    }),
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'B', 'C', 'D'],
                    ['A', BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.WRITEIN],
                    ['A', BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.SKIPPED],
                    ['write-in', 'A', 'C', BallotMarks.OVERVOTE],
                    ['write-in', 'B', 'B', BallotMarks.OVERVOTE],
                    ['C', 'D', BallotMarks.WRITEIN, 'B'],
                    ['C', 'D', 'D', 'B'],
                    ['D', 'B', BallotMarks.WRITEIN, 'D']
                ],
                'precinct': [1, 1, 1, 2, 2, 2, 3, 3]
            },
            'split_fields': ['precinct']
        },
        'expected': {
            'stat': BallotMarks.WRITEIN
        }
    })
]


@pytest.mark.parametrize("param", params)
def test_winner(param):
    rcv = SingleWinner(**param['input'])
    assert rcv.stats()[0]['winner'].item() == param['expected']['stat']


params = [
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'B', 'C', 'D'],
                    ['A', BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.WRITEIN],
                    ['write-in', 'A', 'C', BallotMarks.OVERVOTE],
                    ['write-in', 'B', 'B', BallotMarks.OVERVOTE],
                    ['C', 'A', 'B', 'B']
                ]
            }
        },
        'expected': {
            'stat': 80
        }
    }),
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'B', 'C', 'D'],
                    ['A', BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.WRITEIN],
                    ['write-in', 'A', 'C', BallotMarks.OVERVOTE],
                    ['write-in', 'B', 'B', BallotMarks.OVERVOTE],
                    ['C', 'D', BallotMarks.WRITEIN, 'B']
                ]
            }
        },
        'expected': {
            'stat': 80
        }
    }),
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'B', 'C', 'D'],
                    ['A', BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.WRITEIN],
                    ['A', BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.SKIPPED],
                    ['write-in', 'A', 'C', BallotMarks.OVERVOTE],
                    ['write-in', 'B', 'B', BallotMarks.OVERVOTE],
                    ['C', 'D', BallotMarks.WRITEIN, 'B'],
                    ['C', 'D', 'D', 'B'],
                    ['D', 'B', BallotMarks.WRITEIN, 'D']
                ],
                'precinct': [1, 1, 1, 2, 2, 2, 3, 3]
            },
            'split_fields': ['precinct']
        },
        'expected': {
            'stat': 100 * 5/8
        }
    })
]


@pytest.mark.parametrize("param", params)
def test_winners_consensus_value(param):
    rcv = SingleWinner(**param['input'])
    assert rcv.stats()[0]['winners_consensus_value'].item() == param['expected']['stat']


params = [
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'B', 'C', 'D'],
                    ['A', BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.WRITEIN],
                    ['write-in', 'A', 'C', BallotMarks.OVERVOTE],
                    ['write-in', 'B', 'B', BallotMarks.OVERVOTE],
                    ['C', 'A', 'B', 'B']
                ]
            }
        },
        'expected': {
            'stat': 5
        }
    }),
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'B', 'C', 'D'],
                    ['A', BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.WRITEIN],
                    ['write-in', 'A', 'C', BallotMarks.OVERVOTE],
                    ['write-in', 'B', 'B', BallotMarks.OVERVOTE],
                    ['C', 'D', BallotMarks.WRITEIN, 'B']
                ]
            }
        },
        'expected': {
            'stat': 5
        }
    }),
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'B', 'C', 'D'],
                    ['A', BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.WRITEIN],
                    ['A', BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.SKIPPED],
                    ['write-in', 'A', 'C', BallotMarks.OVERVOTE],
                    ['write-in', 'B', 'B', BallotMarks.OVERVOTE],
                    ['C', 'D', BallotMarks.WRITEIN, 'B'],
                    ['C', 'D', 'D', 'B'],
                    ['D', 'B', BallotMarks.WRITEIN, 'D']
                ]
            }
        },
        'expected': {
            'stat': 8
        }
    }),
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'B', 'C', 'D'],
                    ['A', BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.WRITEIN],
                    ['A', BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.SKIPPED],
                    ['write-in', 'A', 'C', BallotMarks.OVERVOTE],
                    ['write-in', 'B', 'B', BallotMarks.OVERVOTE],
                    ['C', 'D', BallotMarks.WRITEIN, 'B'],
                    ['C', 'D', 'D', 'B'],
                    ['D', 'B', BallotMarks.WRITEIN, 'D']
                ],
                'weight': [1, 1, 1, 2, 2, 1, 1, 1],
                'precinct': [1, 1, 1, 2, 2, 2, 3, 3]
            },
            'split_fields': ['precinct']
        },
        'expected': {
            'stat': 10
        }
    })
]


@pytest.mark.parametrize("param", params)
def test_first_round_active_votes(param):
    rcv = SingleWinner(**param['input'])
    assert rcv.stats()[0]['first_round_active_votes'].item() == param['expected']['stat']


params = [
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'B', 'C', 'D'],
                    ['A', BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.WRITEIN],
                    ['write-in', 'A', 'C', BallotMarks.OVERVOTE],
                    ['write-in', 'B', 'B', BallotMarks.OVERVOTE],
                    ['C', 'A', 'B', 'B']
                ]
            }
        },
        'expected': {
            'stat': 5
        }
    }),
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'B', 'C', 'D'],
                    ['A', BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.WRITEIN],
                    ['write-in', 'A', 'C', BallotMarks.OVERVOTE],
                    ['write-in', 'B', 'B', BallotMarks.OVERVOTE],
                    ['C', 'D', BallotMarks.WRITEIN, 'B']
                ]
            }
        },
        'expected': {
            'stat': 5
        }
    }),
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'B', 'C', 'D'],
                    ['A', BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.WRITEIN],
                    ['A', BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.SKIPPED],
                    ['write-in', 'A', 'C', BallotMarks.OVERVOTE],
                    ['write-in', 'B', 'B', BallotMarks.OVERVOTE],
                    ['C', 'D', BallotMarks.WRITEIN, 'B'],
                    ['C', 'D', 'D', 'B'],
                    ['D', 'B', BallotMarks.WRITEIN, 'D']
                ]
            }
        },
        'expected': {
            'stat': 7
        }
    }),
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'B', 'C', 'D'],
                    ['A', BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.WRITEIN],
                    ['A', BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.SKIPPED],
                    ['write-in', 'A', 'C', BallotMarks.OVERVOTE],
                    ['write-in', 'B', 'B', BallotMarks.OVERVOTE],
                    ['C', 'D', BallotMarks.WRITEIN, 'B'],
                    ['C', 'D', 'D', 'B'],
                    ['D', 'B', BallotMarks.WRITEIN, 'D']
                ],
                'weight': [1, 1, 1, 2, 2, 1, 1, 1],
                'precinct': [1, 1, 1, 2, 2, 2, 3, 3]
            },
            'split_fields': ['precinct']
        },
        'expected': {
            'stat': 9
        }
    })
]


@pytest.mark.parametrize("param", params)
def test_final_round_active_votes(param):
    rcv = SingleWinner(**param['input'])
    assert rcv.stats()[0]['final_round_active_votes'].item() == param['expected']['stat']


params = [
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'B', 'C', 'D'],
                    ['A', BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.WRITEIN],
                    ['write-in', 'A', 'C', BallotMarks.OVERVOTE],
                    ['write-in', 'B', 'B', BallotMarks.OVERVOTE],
                    ['C', 'A', 'B', 'B']
                ]
            }
        },
        'expected': {
            'stat': 0
        }
    }),
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'B', 'C', 'D'],
                    ['A', BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.WRITEIN],
                    ['write-in', 'A', 'C', BallotMarks.OVERVOTE],
                    ['write-in', 'B', 'B', BallotMarks.OVERVOTE],
                    ['C', 'D', BallotMarks.WRITEIN, 'B']
                ]
            }
        },
        'expected': {
            'stat': 0
        }
    }),
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'B', 'C', 'D'],
                    ['A', BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.WRITEIN],
                    ['A', BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.SKIPPED],
                    ['write-in', 'A', 'C', BallotMarks.OVERVOTE],
                    ['write-in', 'B', 'B', BallotMarks.OVERVOTE],
                    ['C', 'D', BallotMarks.WRITEIN, 'B'],
                    ['C', 'D', 'D', 'B'],
                    ['D', 'B', BallotMarks.WRITEIN, 'D'],
                    [BallotMarks.OVERVOTE, BallotMarks.OVERVOTE, BallotMarks.OVERVOTE, BallotMarks.OVERVOTE]
                ]
            }
        },
        'expected': {
            'stat': 1
        }
    }),
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'B', 'C', 'D'],
                    ['A', BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.WRITEIN],
                    ['A', BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.SKIPPED],
                    ['write-in', 'A', 'C', BallotMarks.OVERVOTE],
                    ['write-in', 'B', 'B', BallotMarks.OVERVOTE],
                    ['C', 'D', BallotMarks.WRITEIN, 'B'],
                    ['C', 'D', 'D', 'B'],
                    ['D', 'B', BallotMarks.WRITEIN, 'D'],
                    [BallotMarks.OVERVOTE, BallotMarks.OVERVOTE, BallotMarks.OVERVOTE, BallotMarks.OVERVOTE]
                ],
                'weight': [1, 1, 1, 2, 2, 1, 1, 1, 5],
                'precinct': [1, 1, 1, 2, 2, 2, 3, 3, 4]
            },
            'split_fields': ['precinct']
        },
        'expected': {
            'stat': 5
        }
    })
]


@pytest.mark.parametrize("param", params)
def test_total_pretally_exhausted(param):
    rcv = SingleWinner(**param['input'])
    assert rcv.stats()[0]['total_pretally_exhausted'].item() == param['expected']['stat']


params = [
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'B', 'C', 'D'],
                    ['A', BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.WRITEIN],
                    ['write-in', 'A', 'C', BallotMarks.OVERVOTE],
                    ['write-in', 'B', 'B', BallotMarks.OVERVOTE],
                    ['C', 'A', 'B', 'B']
                ]
            }
        },
        'expected': {
            'stat': 0
        }
    }),
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'B', 'C', 'D'],
                    ['A', BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.WRITEIN],
                    ['write-in', 'A', 'C', BallotMarks.OVERVOTE],
                    ['write-in', 'B', 'B', BallotMarks.OVERVOTE],
                    ['C', 'D', BallotMarks.WRITEIN, 'B']
                ]
            }
        },
        'expected': {
            'stat': 0
        }
    }),
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'B', 'C', 'D'],
                    ['A', BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.WRITEIN],
                    ['A', BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.SKIPPED],
                    ['write-in', 'A', 'C', BallotMarks.OVERVOTE],
                    ['write-in', 'B', 'B', BallotMarks.OVERVOTE],
                    ['C', 'D', BallotMarks.WRITEIN, 'B'],
                    ['C', 'D', 'D', 'B'],
                    ['D', 'B', BallotMarks.WRITEIN, 'D'],
                    [BallotMarks.OVERVOTE, BallotMarks.OVERVOTE, BallotMarks.OVERVOTE, BallotMarks.OVERVOTE]
                ]
            }
        },
        'expected': {
            'stat': 1
        }
    }),
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'C', 'C', 'D'],
                    [BallotMarks.OVERVOTE, 'A', 'B', 'C'],
                    ['B', BallotMarks.OVERVOTE, 'A', 'B'],
                    ['B', 'A', 'A', 'C'],
                    ['C', BallotMarks.SKIPPED, BallotMarks.SKIPPED, 'A'],
                    ['C', 'A', 'B', 'E'],
                    ['D', BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.SKIPPED],
                    ['E', BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.SKIPPED]
                ],
                'weight': [1, 1, 1, 1, 1, 1, 5, 5],
                'precinct': [1, 1, 1, 2, 2, 2, 3, 3]
            },
            'split_fields': ['precinct'],
            'exhaust_on_duplicate_candidate_marks': True,
            'exhaust_on_overvote_marks': True,
            'exhaust_on_repeated_skipped_marks': True
        },
        'expected': {
            'stat': 4
        }
    })
]


@pytest.mark.parametrize("param", params)
def test_total_posttally_exhausted(param):
    rcv = SingleWinner(**param['input'])
    assert rcv.stats()[0]['total_posttally_exhausted'].item() == param['expected']['stat']


params = [
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'C', 'C', 'D'],
                    [BallotMarks.OVERVOTE, 'A', 'B', 'C'],
                    ['B', BallotMarks.OVERVOTE, 'A', 'B'],
                    ['B', 'A', 'A', 'C'],
                    ['C', BallotMarks.SKIPPED, BallotMarks.SKIPPED, 'A'],
                    ['C', 'A', 'B', 'E'],
                    ['D', BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.SKIPPED],
                    ['E', BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.SKIPPED]
                ],
                'weight': [1, 1, 1, 1, 1, 1, 5, 5],
                'precinct': [1, 1, 1, 2, 2, 2, 3, 3]
            },
            'split_fields': ['precinct'],
            'exhaust_on_duplicate_candidate_marks': True,
            'exhaust_on_overvote_marks': True,
            'exhaust_on_repeated_skipped_marks': True
        },
        'expected': {
            'stat': 1
        }
    })
]


@pytest.mark.parametrize("param", params)
def test_total_posttally_exhausted_by_overvote(param):
    rcv = SingleWinner(**param['input'])
    assert rcv.stats()[0]['total_posttally_exhausted_by_overvote'].item() == param['expected']['stat']


params = [
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'C', 'C', 'D'],
                    [BallotMarks.OVERVOTE, 'A', 'B', 'C'],
                    ['B', BallotMarks.OVERVOTE, 'A', 'B'],
                    ['B', 'A', 'A', 'C'],
                    ['C', BallotMarks.SKIPPED, BallotMarks.SKIPPED, 'A'],
                    ['C', 'A', 'B', 'E'],
                    ['D', BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.SKIPPED],
                    ['E', BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.SKIPPED]
                ],
                'weight': [1, 1, 1, 1, 1, 1, 5, 5],
                'precinct': [1, 1, 1, 2, 2, 2, 3, 3]
            },
            'split_fields': ['precinct'],
            'exhaust_on_duplicate_candidate_marks': True,
            'exhaust_on_overvote_marks': True,
            'exhaust_on_repeated_skipped_marks': True
        },
        'expected': {
            'stat': 1
        }
    })
]


@pytest.mark.parametrize("param", params)
def test_total_posttally_exhausted_by_skipped_rankings(param):
    rcv = SingleWinner(**param['input'])
    assert rcv.stats()[0]['total_posttally_exhausted_by_skipped_rankings'].item() == param['expected']['stat']


params = [
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'C', 'C', 'D'],
                    [BallotMarks.OVERVOTE, 'A', 'B', 'C'],
                    ['B', BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.SKIPPED],
                    ['B', 'A', 'A', 'C'],
                    ['C', BallotMarks.SKIPPED, BallotMarks.SKIPPED, 'A'],
                    ['C', 'A', 'B', 'E'],
                    ['D', BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.SKIPPED],
                    ['E', BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.SKIPPED]
                ],
                'weight': [1, 1, 1, 1, 1, 1, 5, 5],
                'precinct': [1, 1, 1, 2, 2, 2, 3, 3]
            },
            'split_fields': ['precinct'],
            'exhaust_on_duplicate_candidate_marks': True,
            'exhaust_on_overvote_marks': True,
            'exhaust_on_repeated_skipped_marks': True
        },
        'expected': {
            'stat': 1
        }
    })
]


@pytest.mark.parametrize("param", params)
def test_total_posttally_exhausted_by_abstention(param):
    rcv = SingleWinner(**param['input'])
    assert rcv.stats()[0]['total_posttally_exhausted_by_abstention'].item() == param['expected']['stat']


params = [
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'C', 'C', 'D'],
                    [BallotMarks.OVERVOTE, 'A', 'B', 'C'],
                    ['B', BallotMarks.OVERVOTE, 'A', 'B'],
                    ['B', 'A', 'A', 'C'],
                    ['C', BallotMarks.SKIPPED, BallotMarks.SKIPPED, 'A'],
                    ['C', 'A', 'B', 'E'],
                    ['D', BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.SKIPPED],
                    ['E', BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.SKIPPED]
                ],
                'weight': [1, 1, 1, 1, 1, 1, 5, 5],
                'precinct': [1, 1, 1, 2, 2, 2, 3, 3]
            },
            'split_fields': ['precinct'],
            'exhaust_on_duplicate_candidate_marks': True,
            'exhaust_on_overvote_marks': True,
            'exhaust_on_repeated_skipped_marks': True
        },
        'expected': {
            'stat': 0
        }
    })
]


@pytest.mark.parametrize("param", params)
def test_total_posttally_exhausted_by_rank_limit(param):
    rcv = SingleWinner(**param['input'])
    assert rcv.stats()[0]['total_posttally_exhausted_by_rank_limit'].item() == param['expected']['stat']


params = [
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'C', 'C', 'D'],
                    [BallotMarks.OVERVOTE, 'A', 'B', 'C'],
                    ['B', BallotMarks.OVERVOTE, 'A', 'B'],
                    ['B', 'A', 'A', 'C'],
                    ['C', BallotMarks.SKIPPED, BallotMarks.SKIPPED, 'A'],
                    ['C', 'A', 'B', 'E'],
                    ['D', BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.SKIPPED],
                    ['E', BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.SKIPPED]
                ],
                'weight': [1, 1, 1, 1, 1, 1, 5, 5],
                'precinct': [1, 1, 1, 2, 2, 2, 3, 3]
            },
            'split_fields': ['precinct'],
            'exhaust_on_duplicate_candidate_marks': True,
            'exhaust_on_overvote_marks': True,
            'exhaust_on_repeated_skipped_marks': True
        },
        'expected': {
            'stat': 0
        }
    })
]


@pytest.mark.parametrize("param", params)
def test_total_posttally_exhausted_by_rank_limit_fully_ranked(param):
    rcv = SingleWinner(**param['input'])
    assert rcv.stats()[0]['total_posttally_exhausted_by_rank_limit_fully_ranked'].item() == param['expected']['stat']


params = [
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'C', 'C', 'D'],
                    [BallotMarks.OVERVOTE, 'A', 'B', 'C'],
                    ['B', BallotMarks.OVERVOTE, 'A', 'B'],
                    ['B', 'A', 'A', 'C'],
                    ['C', BallotMarks.SKIPPED, BallotMarks.SKIPPED, 'A'],
                    ['C', 'A', 'B', 'E'],
                    ['D', BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.SKIPPED],
                    ['E', BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.SKIPPED]
                ],
                'weight': [1, 1, 1, 1, 1, 1, 5, 5],
                'precinct': [1, 1, 1, 2, 2, 2, 3, 3]
            },
            'split_fields': ['precinct'],
            'exhaust_on_duplicate_candidate_marks': True,
            'exhaust_on_overvote_marks': True,
            'exhaust_on_repeated_skipped_marks': True
        },
        'expected': {
            'stat': 0
        }
    })
]


@pytest.mark.parametrize("param", params)
def test_total_posttally_exhausted_by_rank_limit_partially_ranked(param):
    rcv = SingleWinner(**param['input'])
    assert rcv.stats()[0]['total_posttally_exhausted_by_rank_limit_partially_ranked'].item() == param['expected']['stat']


params = [
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'C', 'C', 'D'],
                    [BallotMarks.OVERVOTE, 'A', 'B', 'C'],
                    ['B', BallotMarks.OVERVOTE, 'A', 'B'],
                    ['B', 'A', 'A', 'C'],
                    ['C', BallotMarks.SKIPPED, BallotMarks.SKIPPED, 'A'],
                    ['C', 'A', 'B', 'E'],
                    ['D', BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.SKIPPED],
                    ['E', BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.SKIPPED]
                ],
                'weight': [1, 1, 1, 1, 1, 1, 5, 5],
                'precinct': [1, 1, 1, 2, 2, 2, 3, 3]
            },
            'split_fields': ['precinct'],
            'exhaust_on_duplicate_candidate_marks': True,
            'exhaust_on_overvote_marks': True,
            'exhaust_on_repeated_skipped_marks': True
        },
        'expected': {
            'stat': 2
        }
    })
]


@pytest.mark.parametrize("param", params)
def test_total_posttally_exhausted_by_duplicate_rankings(param):
    rcv = SingleWinner(**param['input'])
    assert rcv.stats()[0]['total_posttally_exhausted_by_duplicate_rankings'].item() == param['expected']['stat']


params = [
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'B', 'C', 'D'],
                    ['A', BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.WRITEIN],
                    ['write-in', 'A', 'C', BallotMarks.OVERVOTE],
                    ['write-in', 'B', 'B', BallotMarks.OVERVOTE],
                    ['C', 'A', 'B', 'B']
                ],
                'split': [1, 1, 2, 2, 2]
            },
            'split_fields': ['split']
        },
        'expected': {
            'stat': [0, 0]
        }
    }),
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'B', 'C', 'D'],
                    ['A', BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.WRITEIN],
                    ['write-in', 'A', 'C', BallotMarks.OVERVOTE],
                    ['write-in', 'B', 'B', BallotMarks.OVERVOTE],
                    ['C', 'D', BallotMarks.WRITEIN, 'B']
                ],
                'split': [1, 1, 2, 2, 2]
            },
            'split_fields': ['split']
        },
        'expected': {
            'stat': [0, 0]
        }
    }),
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'B', 'C', 'D'],
                    ['A', BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.WRITEIN],
                    ['A', BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.SKIPPED],
                    ['write-in', 'A', 'C', BallotMarks.OVERVOTE],
                    ['write-in', 'B', 'B', BallotMarks.OVERVOTE],
                    ['C', 'D', BallotMarks.WRITEIN, 'B'],
                    ['C', 'D', 'D', 'B'],
                    ['D', 'B', BallotMarks.WRITEIN, 'D'],
                    [BallotMarks.OVERVOTE, BallotMarks.OVERVOTE, BallotMarks.OVERVOTE, BallotMarks.OVERVOTE]
                ],
                'split': [1, 1, 1, 1, 2, 2, 2, 2, 2]
            },
            'split_fields': ['split']
        },
        'expected': {
            'stat': [0, 1]
        }
    }),
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'B', 'C', 'D'],
                    ['A', BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.WRITEIN],
                    ['A', BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.SKIPPED],
                    ['write-in', 'A', 'C', BallotMarks.OVERVOTE],
                    ['write-in', 'B', 'B', BallotMarks.OVERVOTE],
                    ['C', 'D', BallotMarks.WRITEIN, 'B'],
                    ['C', 'D', 'D', 'B'],
                    ['D', 'B', BallotMarks.WRITEIN, 'D'],
                    [BallotMarks.OVERVOTE, BallotMarks.OVERVOTE, BallotMarks.OVERVOTE, BallotMarks.OVERVOTE]
                ],
                'weight': [1, 1, 1, 2, 2, 1, 1, 1, 5],
                'split': [1, 1, 1, 1, 2, 2, 2, 2, 2]
            },
            'split_fields': ['split']
        },
        'expected': {
            'stat': [0, 5]
        }
    })
]


@pytest.mark.parametrize("param", params)
def test_split_total_pretally_exhausted(param):
    rcv = SingleWinner(**param['input'])
    assert rcv.stats(add_split_stats=True)[0]['split_total_pretally_exhausted'].tolist() == param['expected']['stat']


params = [
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'B', 'C', 'D'],
                    ['A', BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.WRITEIN],
                    ['write-in', 'A', 'C', BallotMarks.OVERVOTE],
                    ['write-in', 'B', 'B', BallotMarks.OVERVOTE],
                    ['C', 'A', 'B', 'B']
                ],
                'split': [1, 1, 2, 2, 2]
            },
            'split_fields': ['split']
        },
        'expected': {
            'stat': [0, 0]
        }
    }),
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'B', 'C', 'D'],
                    ['A', BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.WRITEIN],
                    ['write-in', 'A', 'C', BallotMarks.OVERVOTE],
                    ['write-in', 'B', 'B', BallotMarks.OVERVOTE],
                    ['C', 'D', BallotMarks.WRITEIN, 'B']
                ],
                'split': [1, 1, 2, 2, 2]
            },
            'split_fields': ['split']
        },
        'expected': {
            'stat': [0, 0]
        }
    }),
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'B', 'C', 'D'],
                    ['A', BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.WRITEIN],
                    ['A', BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.SKIPPED],
                    ['write-in', 'A', 'C', BallotMarks.OVERVOTE],
                    ['write-in', 'B', 'B', BallotMarks.OVERVOTE],
                    ['C', 'D', BallotMarks.WRITEIN, 'B'],
                    ['C', 'D', 'D', 'B'],
                    ['D', 'B', BallotMarks.WRITEIN, 'D'],
                    [BallotMarks.OVERVOTE, BallotMarks.OVERVOTE, BallotMarks.OVERVOTE, BallotMarks.OVERVOTE]
                ],
                'split': [1, 1, 1, 1, 2, 2, 2, 2, 2]
            },
            'split_fields': ['split']
        },
        'expected': {
            'stat': [0, 1]
        }
    }),
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'B', 'C', 'D'],
                    ['A', BallotMarks.SKIPPED, BallotMarks.OVERVOTE, BallotMarks.WRITEIN],
                    ['A', BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.SKIPPED],
                    ['write-in', 'A', 'C', BallotMarks.OVERVOTE],
                    ['write-in', 'B', 'B', BallotMarks.OVERVOTE],
                    ['C', 'D', BallotMarks.WRITEIN, 'B'],
                    ['C', 'D', 'D', 'B'],
                    ['D', 'B', BallotMarks.WRITEIN, 'D'],
                    [BallotMarks.OVERVOTE, BallotMarks.OVERVOTE, BallotMarks.OVERVOTE, BallotMarks.OVERVOTE]
                ],
                'weight': [1, 1, 1, 2, 2, 1, 1, 1, 5],
                'split': [1, 1, 1, 1, 2, 2, 2, 2, 2]
            },
            'split_fields': ['split']
        },
        'expected': {
            'stat': [0, 1]
        }
    })
]


@pytest.mark.parametrize("param", params)
def test_split_total_posttally_exhausted(param):
    rcv = SingleWinner(**param['input'])
    assert rcv.stats(add_split_stats=True)[0]['split_total_posttally_exhausted'].tolist() == param['expected']['stat']


params = [
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'C', 'C', 'D'],
                    [BallotMarks.OVERVOTE, 'A', 'B', 'C'],
                    ['B', BallotMarks.OVERVOTE, 'A', 'B'],
                    ['B', 'A', 'A', 'C'],
                    ['C', BallotMarks.SKIPPED, BallotMarks.SKIPPED, 'A'],
                    ['C', 'A', 'B', 'E'],
                    ['D', BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.SKIPPED],
                    ['E', BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.SKIPPED]
                ],
                'weight': [1, 1, 1, 1, 1, 1, 5, 5],
                'split': [1, 1, 1, 1, 2, 2, 2, 2]
            },
            'split_fields': ['split'],
            'exhaust_on_duplicate_candidate_marks': True,
            'exhaust_on_overvote_marks': True,
            'exhaust_on_repeated_skipped_marks': True
        },
        'expected': {
            'stat': [1, 0]
        }
    })
]


@pytest.mark.parametrize("param", params)
def test_split_total_posttally_exhausted_by_overvote(param):
    rcv = SingleWinner(**param['input'])
    assert rcv.stats(add_split_stats=True)[0]['split_total_posttally_exhausted_by_overvote'].tolist() == param['expected']['stat']


params = [
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'C', 'C', 'D'],
                    [BallotMarks.OVERVOTE, 'A', 'B', 'C'],
                    ['B', BallotMarks.OVERVOTE, 'A', 'B'],
                    ['B', 'A', 'A', 'C'],
                    ['C', BallotMarks.SKIPPED, BallotMarks.SKIPPED, 'A'],
                    ['C', 'A', 'B', 'E'],
                    ['D', BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.SKIPPED],
                    ['E', BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.SKIPPED]
                ],
                'weight': [1, 1, 1, 1, 1, 1, 5, 5],
                'split': [1, 1, 1, 1, 2, 2, 2, 2]
            },
            'split_fields': ['split'],
            'exhaust_on_duplicate_candidate_marks': True,
            'exhaust_on_overvote_marks': True,
            'exhaust_on_repeated_skipped_marks': True
        },
        'expected': {
            'stat': [0, 1]
        }
    })
]


@pytest.mark.parametrize("param", params)
def test_split_total_posttally_exhausted_by_skipped_rankings(param):
    rcv = SingleWinner(**param['input'])
    assert rcv.stats(add_split_stats=True)[0]['split_total_posttally_exhausted_by_skipped_rankings'].tolist() == param['expected']['stat']


params = [
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'C', 'C', 'D'],
                    [BallotMarks.OVERVOTE, 'A', 'B', 'C'],
                    ['B', BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.SKIPPED],
                    ['B', 'A', 'A', 'C'],
                    ['C', BallotMarks.SKIPPED, BallotMarks.SKIPPED, 'A'],
                    ['C', 'A', 'B', 'E'],
                    ['D', BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.SKIPPED],
                    ['E', BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.SKIPPED]
                ],
                'weight': [1, 1, 1, 1, 1, 1, 5, 5],
                'split': [1, 1, 1, 1, 2, 2, 2, 2]
            },
            'split_fields': ['split'],
            'exhaust_on_duplicate_candidate_marks': True,
            'exhaust_on_overvote_marks': True,
            'exhaust_on_repeated_skipped_marks': True
        },
        'expected': {
            'stat': [1, 0]
        }
    })
]


@pytest.mark.parametrize("param", params)
def test_split_total_posttally_exhausted_by_abstention(param):
    rcv = SingleWinner(**param['input'])
    assert rcv.stats(add_split_stats=True)[0]['split_total_posttally_exhausted_by_abstention'].tolist() == param['expected']['stat']


params = [
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'C', 'C', 'D'],
                    [BallotMarks.OVERVOTE, 'A', 'B', 'C'],
                    ['B', BallotMarks.OVERVOTE, 'A', 'B'],
                    ['B', 'A', 'A', 'C'],
                    ['C', BallotMarks.SKIPPED, BallotMarks.SKIPPED, 'A'],
                    ['C', 'A', 'B', 'E'],
                    ['D', BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.SKIPPED],
                    ['E', BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.SKIPPED]
                ],
                'weight': [1, 1, 1, 1, 1, 1, 5, 5],
                'split': [1, 1, 1, 1, 2, 2, 2, 3]
            },
            'split_fields': ['split'],
            'exhaust_on_duplicate_candidate_marks': True,
            'exhaust_on_overvote_marks': True,
            'exhaust_on_repeated_skipped_marks': True
        },
        'expected': {
            'stat': [0, 0, 0]
        }
    })
]


@pytest.mark.parametrize("param", params)
def test_split_total_posttally_exhausted_by_rank_limit(param):
    rcv = SingleWinner(**param['input'])
    assert rcv.stats(add_split_stats=True)[0]['split_total_posttally_exhausted_by_rank_limit'].tolist() == param['expected']['stat']


params = [
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'C', 'C', 'D'],
                    [BallotMarks.OVERVOTE, 'A', 'B', 'C'],
                    ['B', BallotMarks.OVERVOTE, 'A', 'B'],
                    ['B', 'A', 'A', 'C'],
                    ['C', BallotMarks.SKIPPED, BallotMarks.SKIPPED, 'A'],
                    ['C', 'A', 'B', 'E'],
                    ['D', BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.SKIPPED],
                    ['E', BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.SKIPPED]
                ],
                'weight': [1, 1, 1, 1, 1, 1, 5, 5],
                'split': [1, 1, 1, 1, 2, 2, 2, 3]
            },
            'split_fields': ['split'],
            'exhaust_on_duplicate_candidate_marks': True,
            'exhaust_on_overvote_marks': True,
            'exhaust_on_repeated_skipped_marks': True
        },
        'expected': {
            'stat': [0, 0, 0]
        }
    })
]


@pytest.mark.parametrize("param", params)
def test_split_total_posttally_exhausted_by_rank_limit_fully_ranked(param):
    rcv = SingleWinner(**param['input'])
    assert rcv.stats(add_split_stats=True)[0]['split_total_posttally_exhausted_by_rank_limit_fully_ranked'].tolist() == param['expected']['stat']


params = [
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'C', 'C', 'D'],
                    [BallotMarks.OVERVOTE, 'A', 'B', 'C'],
                    ['B', BallotMarks.OVERVOTE, 'A', 'B'],
                    ['B', 'A', 'A', 'C'],
                    ['C', BallotMarks.SKIPPED, BallotMarks.SKIPPED, 'A'],
                    ['C', 'A', 'B', 'E'],
                    ['D', BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.SKIPPED],
                    ['E', BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.SKIPPED]
                ],
                'weight': [1, 1, 1, 1, 1, 1, 5, 5],
                'split': [1, 1, 1, 1, 2, 2, 2, 3]
            },
            'split_fields': ['split'],
            'exhaust_on_duplicate_candidate_marks': True,
            'exhaust_on_overvote_marks': True,
            'exhaust_on_repeated_skipped_marks': True
        },
        'expected': {
            'stat': [0, 0, 0]
        }
    })
]


@pytest.mark.parametrize("param", params)
def test_split_total_posttally_exhausted_by_rank_limit_partially_ranked(param):
    rcv = SingleWinner(**param['input'])
    assert rcv.stats(add_split_stats=True)[0]['split_total_posttally_exhausted_by_rank_limit_partially_ranked'].tolist() == param['expected']['stat']


params = [
    ({
        'input': {
            'parsed_cvr': {
                'ranks': [
                    ['A', 'C', 'C', 'D'],
                    [BallotMarks.OVERVOTE, 'A', 'B', 'C'],
                    ['B', BallotMarks.OVERVOTE, 'A', 'B'],
                    ['B', 'A', 'A', 'C'],
                    ['C', BallotMarks.SKIPPED, BallotMarks.SKIPPED, 'A'],
                    ['C', 'A', 'B', 'E'],
                    ['D', BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.SKIPPED],
                    ['E', BallotMarks.SKIPPED, BallotMarks.SKIPPED, BallotMarks.SKIPPED]
                ],
                'weight': [1, 1, 1, 1, 1, 1, 5, 5],
                'split': [1, 1, 1, 1, 2, 2, 2, 2]
            },
            'split_fields': ['split'],
            'exhaust_on_duplicate_candidate_marks': True,
            'exhaust_on_overvote_marks': True,
            'exhaust_on_repeated_skipped_marks': True
        },
        'expected': {
            'stat': [2, 0]
        }
    })
]


@pytest.mark.parametrize("param", params)
def test_split_total_posttally_exhausted_by_duplicate_rankings(param):
    rcv = SingleWinner(**param['input'])
    assert rcv.stats(add_split_stats=True)[0]['split_total_posttally_exhausted_by_duplicate_rankings'].tolist() == param['expected']['stat']
